"""
Personality Test
Copyright (c) 2022 Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 6/20/2022
Updated: 6/21/2022
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Disclaimer:

    This test is provided for educational and entertainment uses only.
    The test is not clinically administered and as such the results are
    not suitable for aiding important decisions.
    The test is also fallible, so, if the results say something about
    you that you don't think is true, then you are right and it is wrong.

References:

    - Administering IPIP Measures, with a 50-item Sample Questionnaire
    URL: <https://ipip.ori.org/new_ipip-50-item-scale.htm>

"""
# Standard imports.
import os

# External imports.
from dotenv import dotenv_values
from cannlytics.firebase import initialize_firebase, update_document
from cannlytics.utils import snake_case


PROMPT = """Describe yourself as you generally are now, not as you wish
to be in the future. Describe yourself as you honestly see yourself, in
relation to other people you know of the same sex as you are, and
roughly your same age. So that you can describe yourself in an honest
manner, your responses will be kept in absolute confidence. Indicate for
each statement whether it is

1. Very Inaccurate,
2. Moderately Inaccurate,
3. Neither Accurate Nor Inaccurate,
4. Moderately Accurate, or
5. Very Accurate

as a description of you.
"""

DISCLAIMER = """This test is provided for educational and entertainment uses only.
The test is not clinically administered and as such the results are
not suitable for aiding important decisions.
The test is also fallible, so, if the results say something about
you that you don't think is true, then you are right and it is wrong.
"""

POSITIVE_SCALE = {
    '1': 'Very Inaccurate',
    '2': 'Moderately Inaccurate',
    '3': 'Neither Accurate Nor Inaccurate',
    '4': 'Moderately Accurate',
    '5': 'Very Accurate',
}

NEGATIVE_SCALE = {
    '5': 'Very Inaccurate',
    '4': 'Moderately Inaccurate',
    '3': 'Neither Accurate Nor Inaccurate',
    '2': 'Moderately Accurate',
    '1': 'Very Accurate',
}

FACTORS = {
    '1': 'Extraversion',
    '2': 'Agreeableness',
    '3': 'Conscientiousness',
    '4': 'Neuroticism',
    '5': 'Openness',
}

MAXES = {
    '1': 5 * 5 - (5 * 1),
    '2': 6 * 5 - (4 * 1),
    '3': 6 * 5 - (4 * 1),
    '4': 2 * 5 - (8 * 1),
    '5': 7 * 5 - (3 * 1),
}

MINS = {
    '1': 5 * 1 - (5 * 5),
    '2': 6 * 1 - (4 * 5),
    '3': 6 * 1 - (4 * 5),
    '4': 2 * 1 - (8 * 5),
    '5': 7 * 1 - (3 * 5),
}

QUESTIONS = [
    {'id': '1', 'factor': 1, 'positive': True, 'text': 'Am the life of the party.', 'key': 'EXT1'},
    {'id': '2', 'factor': 2, 'positive': False, 'text': 'Feel little concern for others.', 'key': 'AGR1'},
    {'id': '3', 'factor': 3, 'positive': True, 'text': 'Am always prepared.', 'key': 'CSN1'},
    {'id': '4', 'factor': 4, 'positive': False, 'text': 'Get stressed out easily.', 'key': 'EST1'},
    {'id': '5', 'factor': 5, 'positive': True, 'text': 'Have a rich vocabulary.', 'key': 'OPN1'},
    {'id': '6', 'factor': 1, 'positive': False, 'text': "Don't talk a lot.", 'key': 'EXT2'},
    {'id': '7', 'factor': 2, 'positive': True, 'text': 'Am interested in people.', 'key': 'AGR2'},
    {'id': '8', 'factor': 3, 'positive': False, 'text': 'Leave my belongings around.', 'key': 'CSN2'},
    {'id': '9', 'factor': 4, 'positive': True, 'text': 'Am relaxed most of the time.', 'key': 'EST2'},
    {'id': '10', 'factor': 5, 'positive': False, 'text': 'Have difficulty understanding abstract ideas.', 'key': 'OPN2'},
    {'id': '11', 'factor': 1, 'positive': True, 'text': 'Feel comfortable around people.', 'key': 'EXT3'},
    {'id': '12', 'factor': 2, 'positive': False, 'text': 'Insult people.', 'key': 'AGR3'},
    {'id': '13', 'factor': 3, 'positive': True, 'text': 'Pay attention to details.', 'key': 'CSN3'},
    {'id': '14', 'factor': 4, 'positive': False, 'text': 'Worry about things.', 'key': 'EST3'},
    {'id': '15', 'factor': 5, 'positive': True, 'text': 'Have a vivid imagination.', 'key': 'OPN3'},
    {'id': '16', 'factor': 1, 'positive': False, 'text': 'Keep in the background.', 'key': 'EXT4'},
    {'id': '17', 'factor': 2, 'positive': True, 'text': "Sympathize with others' feelings.", 'key': 'AGR4'},
    {'id': '18', 'factor': 3, 'positive': False, 'text': 'Make a mess of things.', 'key': 'CSN4'},
    {'id': '19', 'factor': 4, 'positive': True, 'text': 'Seldom feel blue.', 'key': 'EST4'},
    {'id': '20', 'factor': 5, 'positive': False, 'text': 'Am not interested in abstract ideas.', 'key': 'OPN4'},
    {'id': '21', 'factor': 1, 'positive': True, 'text': 'Start conversations.', 'key': 'EXT5'},
    {'id': '22', 'factor': 2, 'positive': False, 'text': "Am not interested in other people's problems.", 'key': 'AGR5'},
    {'id': '23', 'factor': 3, 'positive': True, 'text': 'Get chores done right away.', 'key': 'CSN5'},
    {'id': '24', 'factor': 4, 'positive': False, 'text': 'Am easily disturbed.', 'key': 'EST5'},
    {'id': '25', 'factor': 5, 'positive': True, 'text': 'Have excellent ideas.', 'key': 'OPN5'},
    {'id': '26', 'factor': 1, 'positive': False, 'text': 'Have little to say.', 'key': 'EXT6'},
    {'id': '27', 'factor': 2, 'positive': True, 'text': 'Have a soft heart.', 'key': 'AGR6'},
    {'id': '28', 'factor': 3, 'positive': False, 'text': 'Often forget to put things back in their proper place.', 'key': 'CSN6'},
    {'id': '29', 'factor': 4, 'positive': False, 'text': 'Get upset easily.', 'key': 'EST6'},
    {'id': '30', 'factor': 5, 'positive': False, 'text': 'Do not have a good imagination.', 'key': 'OPN6'},
    {'id': '31', 'factor': 1, 'positive': True, 'text': 'Talk to a lot of different people at parties.', 'key': 'EXT7'},
    {'id': '32', 'factor': 2, 'positive': False, 'text': 'Am not really interested in others.', 'key': 'AGR7'},
    {'id': '33', 'factor': 3, 'positive': True, 'text': 'Like order.', 'key': 'CSN7'},
    {'id': '34', 'factor': 4, 'positive': False, 'text': 'Change my mood a lot.', 'key': 'EST7'},
    {'id': '35', 'factor': 5, 'positive': True, 'text': 'Am quick to understand things.', 'key': 'OPN7'},
    {'id': '36', 'factor': 1, 'positive': False, 'text': "Don't like to draw attention to myself.", 'key': 'EXT8'},
    {'id': '37', 'factor': 2, 'positive': True, 'text': 'Take time out for others.', 'key': 'AGR8'},
    {'id': '38', 'factor': 3, 'positive': False, 'text': 'Shirk my duties.', 'key': 'CSN8'},
    {'id': '39', 'factor': 4, 'positive': False, 'text': 'Have frequent mood swings.', 'key': 'EST8'},
    {'id': '40', 'factor': 5, 'positive': True, 'text': 'Use difficult words.', 'key': 'OPN8'},
    {'id': '41', 'factor': 1, 'positive': True, 'text': "Don't mind being the center of attention.", 'key': 'EXT9'},
    {'id': '42', 'factor': 2, 'positive': True, 'text': "Feel others' emotions.", 'key': 'AGR9'},
    {'id': '43', 'factor': 3, 'positive': True, 'text': 'Follow a schedule.', 'key': 'CSN9'},
    {'id': '44', 'factor': 4, 'positive': False, 'text': 'Get irritated easily.', 'key': 'EST9'},
    {'id': '45', 'factor': 5, 'positive': True, 'text': 'Spend time reflecting on things.', 'key': 'OPN9'},
    {'id': '46', 'factor': 1, 'positive': False, 'text': 'Am quiet around strangers.', 'key': 'EXT10'},
    {'id': '47', 'factor': 2, 'positive': True, 'text': 'Make people feel at ease.', 'key': 'AGR10'},
    {'id': '48', 'factor': 3, 'positive': True, 'text': 'Am exacting in my work.', 'key': 'CSN10'},
    {'id': '49', 'factor': 4, 'positive': False, 'text': 'Often feel blue.', 'key': 'EST10'},
    {'id': '50', 'factor': 5, 'positive': True, 'text': 'Am full of ideas.', 'key': 'OPN10'},
]


def score_personality_test(data):
    """Score a personality test for the "Big 5" personality traits.
    Normalizes the scores from the range of possible scores.
    """
    scores, normalized = {}, {}
    for factor in FACTORS: scores[factor] = 0
    for question in QUESTIONS:
        _id = question['id']
        value = int(data[_id])
        factor = str(question['factor'])
        if question['positive']:
            scores[factor] += value
        else:
            scores[factor] -= value
    for factor, value in scores.items():
        maximum = MAXES[factor]
        minimum = MINS[factor]
        normalized[factor] = (value - minimum) / (maximum - minimum)
    return {snake_case(FACTORS[k]):round(normalized[k], 2) for k in normalized}


if __name__ == '__main__':

    # Test the personality test.
    print('Testing personality test....')
    test = {
        '1': 3,
        '2': 3,
        '3': 3,
        '4': 3,
        '5': 3,
        '6': 3,
        '7': 3,
        '8': 3,
        '9': 3,
        '10': 3,
        '11': 3,
        '12': 3,
        '13': 3,
        '14': 3,
        '15': 3,
        '16': 3,
        '17': 3,
        '18': 3,
        '19': 3,
        '20': 3,
        '21': 3,
        '22': 3,
        '23': 3,
        '24': 3,
        '25': 3,
        '26': 3,
        '27': 3,
        '28': 3,
        '29': 3,
        '30': 3,
        '31': 3,
        '32': 3,
        '33': 3,
        '34': 3,
        '35': 3,
        '36': 3,
        '37': 3,
        '38': 3,
        '39': 3,
        '40': 3,
        '41': 3,
        '42': 3,
        '43': 3,
        '44': 3,
        '45': 3,
        '46': 3,
        '47': 3,
        '48': 3,
        '49': 3,
        '50': 3,
    }
    score = score_personality_test(test)
    for factor, value in score.items():
        assert value == 0.5
    print('Neutral score works.')

    # Initialize Firebase
    config = dotenv_values('../../.env')
    os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = config['GOOGLE_APPLICATION_CREDENTIALS']
    db = initialize_firebase()

    # Upload the constants to Firestore.
    constants = {
        'prompt': PROMPT,
        'disclaimer': DISCLAIMER,
        'factors': FACTORS,
        'maxes': MAXES,
        'mins': MINS,
        'positive_scale': POSITIVE_SCALE,
        'negative_scale': NEGATIVE_SCALE,
        'questions': QUESTIONS,
    }
    ref = 'public/data/variables/personality_test'
    update_document(ref, constants, database=db)
 


"""
Cannlytics Utilities Initialization | Cannlytics
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/6/2021
Updated: 8/18/2022
"""
from .utils import (
    camelcase,
    camel_to_snake,
    clean_column_strings,
    clean_dictionary,
    clean_nested_dictionary,
    convert_to_numeric,
    download_file_from_url,
    decode_pdf,
    dump_column,
    encode_pdf,
    get_date_range,
    get_directory_files,
    get_keywords,
    get_random_string,
    get_timestamp,
    kebab_case,
    remove_dict_fields,
    remove_dict_nulls,
    snake_case,
    update_dict,
    convert_month_year_to_date,
    end_of_month,
    end_of_year,
    end_of_period_timeseries,
    months_elapsed,
    reverse_dataframe,
    set_training_period,
    format_billions,
    format_millions,
    format_thousands,
    format_iso_date,
    sandwich_list,
    sorted_nicely,
    strip_whitespace,
    to_excel_with_style,
    rmerge,
    decode_pdf,
    encode_pdf,
    get_number_of_lines,
    unzip_files,
    reorder_columns,
)

__all__ = [
    'constants',
    'camelcase',
    'camel_to_snake',
    'clean_column_strings',
    'clean_dictionary',
    'clean_nested_dictionary',
    'convert_month_year_to_date',
    'convert_to_numeric',
    'download_file_from_url',
    'dump_column',
    'get_date_range',
    'get_directory_files',
    'get_keywords',
    'get_random_string',
    'get_timestamp',
    'remove_dict_fields',
    'remove_dict_nulls',
    'sandwich_list',
    'snake_case',
    'update_dict',
    'kebab_case',
    'end_of_month',
    'end_of_year',
    'end_of_period_timeseries',
    'months_elapsed',
    'reverse_dataframe',
    'set_training_period',
    'format_billions',
    'format_millions',
    'format_thousands',
    'format_iso_date',
    'sorted_nicely',
    'strip_whitespace',
    'to_excel_with_style',
    'rmerge',
    'decode_pdf',
    'encode_pdf',
    'get_number_of_lines',
    'unzip_files',
    'get_number_of_lines',
    'decode_pdf',
    'encode_pdf',
    'reorder_columns',
]



"""
Metrc Models | Cannlytics
Copyright (c) 2021-2023 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/5/2021
Updated: 1/17/2023
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

This module contains common Metrc models.
"""
# Standard imports:
from typing import Any, Callable, Optional

# Internal imports.
from ..firebase import get_document, update_document
from ..utils.utils import (
    camelcase,
    camel_to_snake,
    clean_dictionary,
    clean_nested_dictionary,
    get_timestamp,
    remove_dict_fields,
    remove_dict_nulls,
    update_dict,
)


class Model(object):
    """Base class for all Metrc models."""

    def __init__(
            self,
            client: Any,
            context: dict,
            license_number: Optional[str] = '',
            function: Optional[Callable] = camel_to_snake
        ):
        """Initialize the model, setting keys as properties."""
        self.client = client
        self._license = license_number
        properties = clean_nested_dictionary(context, function)
        for key in properties:
            self.__dict__[key] = properties[key]

    def __getattr__(self, key):
        return self.__dict__[key]

    def __setattr__(self, key, value):
        self.__dict__[key] = value

    @property
    def uid(self):
        """The model's unique ID."""
        return self.__dict__.get('id')

    @classmethod
    def from_dict(cls, client: Any, context: dict):
        """Initiate a class instance from a dictionary."""
        obj = cls(client, context)
        return obj

    @classmethod
    def from_fb(cls, client: Any, ref: str) -> Any:
        """Initialize a class from Firebase data.
        Args:
            client (Client): A Metrc client instance.
            ref (str): The reference to the document in Firestore.
        Returns:
            (Model): A Metrc model.
        """
        data = get_document(ref)
        obj = cls(client, data)
        return obj

    def to_dict(self) -> dict:
        """Returns the model's properties as a dictionary."""
        data = vars(self).copy()
        [data.pop(x, None) for x in ['_license', 'client', '__class__']]
        return data

    def to_fb(self, ref: Optional[str] = '', col: Optional[str] = ''):
        """Upload the model's properties as a dictionary to Firestore.
        Args:
            ref (str): The Firestore document reference.
            col (str): A Firestore collection, with the UID as document ID.
        """
        data = vars(self).copy()
        [data.pop(x, None) for x in ['_license', 'client']]
        if col:
            update_document(f'{col}/{self.uid}', data)
        else:
            update_document(ref, data)


class Delivery(Model):
    """A class that represents a cannabis home delivery. Sales are reported to
    record the transfer of cannabis products to a consumer, patient or
    caregiver.

    When you request receipts you receive the following object.
    ```js
    {
        "Id": 1,
        "ReceiptNumber": null,
        "SalesDateTime": "2016-01-01T17:35:45.000",
        "SalesCustomerType": "Consumer",
        "PatientLicenseNumber": null,
        "CaregiverLicenseNumber": null,
        "IdentificationMethod": null,
        "TotalPackages": 0,
        "TotalPrice": 0.0,
        "Transactions": [],
        "IsFinal": false,
        "ArchivedDate": null,
        "RecordedDateTime": "0001-01-01T00:00:00+00:00",
        "RecordedByUserName": null,
        "LastModified": "0001-01-01T00:00:00+00:00"
    }
    ```

    When you create a receipt, you pass the following object.
    ```js
    {
        "SalesDateTime": "2016-10-04T16:44:53.000",
        "SalesCustomerType": "Consumer",
        "PatientLicenseNumber": null,
        "CaregiverLicenseNumber": null,
        "IdentificationMethod": null,
        "Transactions": [
            {
                "PackageLabel": "ABCDEF012345670000010331",
                "Quantity": 1.0,
                "UnitOfMeasure": "Ounces",
                "TotalAmount": 9.99
            }
        ]
    }
    ```
    """

    def create(self):
        """Create a receipt record in Metrc."""
        context = self.to_dict()
        data = clean_nested_dictionary(context, camelcase)
        self.client.create_receipts([data], license_number=self._license)

    def update(self, **kwargs):
        """Update the receipt given parameters as keyword arguments."""
        context = self.to_dict().copy()
        data = update_dict(context, **kwargs)
        data = remove_dict_nulls(data)
        self.client.update_receipts([data], self._license)

    def delete(self):
        """Delete the receipt."""
        self.client.delete_receipt(self.id, self._license)


class Category(Model):
    """A class representing an item category.
    ```js
        {
            "Name": "Buds",
            "ProductCategoryType": "Buds",
            "QuantityType": "WeightBased",
            "RequiresStrain": true,
            "RequiresItemBrand": false,
            "RequiresAdministrationMethod": false,
            "RequiresUnitCbdPercent": false,
            "RequiresUnitCbdContent": false,
            "RequiresUnitCbdContentDose": false,
            "RequiresUnitThcPercent": false,
            "RequiresUnitThcContent": false,
            "RequiresUnitThcContentDose": false,
            "RequiresUnitVolume": false,
            "RequiresUnitWeight": false,
            "RequiresServingSize": false,
            "RequiresSupplyDurationDays": false,
            "RequiresNumberOfDoses": false,
            "RequiresPublicIngredients": false,
            "RequiresDescription": false,
            "RequiresProductPhotos": 0,
            "RequiresLabelPhotos": 0,
            "RequiresPackagingPhotos": 0,
            "CanContainSeeds": true,
            "CanBeRemediated": true,
            "CanBeDestroyed": false
        }
    ```
    """
    pass


class Employee(Model):
    """An organization's employee or team member.
    ```js
        {
            "FullName": "Keegan Skeate",
            "License": null
        }
    ```
    """
    pass


class Facility(Model):
    """A Facility represents a building licensed for the growing, processing,
    and/or selling of product. Facilities are created and have their
    permissions determined by a state.
    ```js
    {
        "HireDate": "0001-01-01",
        "IsOwner": false,
        "IsManager": true,
        "Occupations": [],
        "Name": "Cultivation LLC",
        "Alias": "Cultivation on Road St",
        "DisplayName": "Cultivation on Road St",
        "CredentialedDate": "1969-08-15",
        "SupportActivationDate": null,
        "SupportExpirationDate": null,
        "SupportLastPaidDate": null,
        "FacilityType": null,
        "License": {
            "Number": "403-X0001",
            "StartDate": "2013-06-28",
            "EndDate": "2015-12-28",
            "LicenseType": "Medical Cultivation"
        }
    }
    ```
    """

    @property
    def license_number(self):
        """The facilities license number."""
        return self.license['number']

    def get_locations(self, uid='', action=''):
        """Get locations at the facility.
        Args:
            uid (str): The UID of a location, takes precedent over action.
            action (str): A specific filter to apply: `active` or `types`.
        """
        response = self.client.get_locations(
            uid=uid,
            action=action,
            license_number=self.license_number
        )
        return response

    def create_location(self, name, location_type='default'):
        """Create a location at the facility.
        Args:
            name (str): A location name.
            location_type (str): An optional location type:
                `default`, `planting`, or `packing`.
                `default` is assigned by default.
        """
        return self.client.create_locations(
            [name],
            [location_type],
            self.license_number
        )
        # TODO: Implement return_obs

    def create_locations(self, names, types=[]):
        """Create locations at the facility.
        Args:
            names (list): A list of location names.
            types (list): An optional list of location types:
                `default`, `planting`, or `packing`.
                `default` is assigned by default.
        """
        return self.client.create_locations(names, types, self.license_number)
        # TODO: Implement return_obs

    def update_locations(self, ids, names, types=[]):
        """Update locations at the facility.
        Args:
            ids (list): A list of location IDs.
            names (list): A list of location names.
            types (list): A list of location types:
                `default`, `planting`, or `packing`.
        """
        data = []
        for index, location_id in enumerate(ids):
            try:
                location_type = types[index]
            except IndexError:
                location_type = 'Default Location Type'
            data.append({
                'Id': location_id,
                'Name': names[index],
                'LocationTypeName': location_type
            })
        response = self.client.update_locations(
            data,
            license_number=self.license_number
        )
        return response
        # TODO: Implement return_obs

    def delete_location(self, uid):
        """Delete a location at the facility.
        Args:
            uid (str): The UID of a location to delete.
        """
        response = self.client.delete_location(
            uid,
            license_number=self.license_number
        )
        return response


class Item(Model):
    """Items are used to track a licensee's inventory at a given facility.
    Metrc documentation states:

    Items belong to a single facility. Each item has a unique item name,
    category, and strain. Item Names are used for identification, so an item
    name should not simply be a category name. Item names are specific to the
    item in that package or production batch. Categories are pre-defined. The
    item name identifies what is in the package and categories are used for
    grouping similar items for reporting purposes. An item will retain its name
    unless it is re-packaged.
    ```js
        {
            "Id": 1,
            "Name": "Buds",
            "ProductCategoryName": "Buds",
            "ProductCategoryType": "Buds",
            "QuantityType": "WeightBased",
            "DefaultLabTestingState": "NotSubmitted",
            "UnitOfMeasureName": "Ounces",
            "ApprovalStatus": "Approved",
            "ApprovalStatusDateTime": "0001-01-01T00:00:00+00:00",
            "StrainId": 1,
            "StrainName": "Spring Hill Kush",
            "AdministrationMethod": null,
            "UnitCbdPercent": null,
            "UnitCbdContent": null,
            "UnitCbdContentUnitOfMeasureName": null,
            "UnitCbdContentDose": null,
            "UnitCbdContentDoseUnitOfMeasureName": null,
            "UnitThcPercent": null,
            "UnitThcContent": null,
            "UnitThcContentUnitOfMeasureName": null,
            "UnitThcContentDose": null,
            "UnitThcContentDoseUnitOfMeasureName": null,
            "UnitVolume": null,
            "UnitVolumeUnitOfMeasureName": null,
            "UnitWeight": null,
            "UnitWeightUnitOfMeasureName": null,
            "ServingSize": null,
            "SupplyDurationDays": null,
            "NumberOfDoses": null,
            "UnitQuantity": null,
            "UnitQuantityUnitOfMeasureName": null,
            "PublicIngredients": null,
            "Description": null,
            "IsUsed": false
        }
    ```
    """

    RETURNED_VALUES = {
        'ProductCategoryName': 'item_category',
        'StrainName': 'strain',
        'UnitOfMeasureName': 'unit_of_measure',
        'QuantityType': 'quantity_type',
        'DefaultLabTestingState': 'default_lab_testing_state',
        'ApprovalStatus': 'approval_status',
        'ApprovalStatusDateTime': 'approval_status_date_time',
        'StrainId': 'strain_id',
        'AdministrationMethod': 'administration_method',
        'UnitQuantity': 'unit_quantity',
        'UnitQuantityUnitOfMeasureName': 'unit_quantity_unit_of_measure_name',
    }

    def __init__(self, client, properties, license_number=''):
        super().__init__(client, properties, license_number)
        for k, v in self.RETURNED_VALUES.items():
            try:
                self.__dict__[v] = properties[k]
            except KeyError:
                pass

    def create(self, license_number='', return_obs=False):
        """Create an item record in Metrc."""
        context = self.to_dict()
        data = clean_dictionary(context, camelcase)
        self.client.create_items([data], license_number, return_obs=return_obs)

    def update(self, **kwargs):
        """Update the item given parameters as keyword arguments."""
        context = self.to_dict().copy()
        data = update_dict(context, **kwargs)
        data = remove_dict_fields(data, self.RETURNED_VALUES.keys())
        data = remove_dict_nulls(data)
        self.client.update_items([data], self._license)

    def delete(self):
        """Delete the item."""
        self.client.delete_item(self.id, self._license)


LOCATION_FIELDS = {
    'name': 'Name',
    'location_type': 'LocationTypeName',
    'batches': 'ForPlantBatches',
    'plants': 'ForPlants',
    'harvests': 'ForHarvests',
    'packages': 'ForPackages'
}

class Location(Model):
    """A class that represents a cannabis-production location.
    ```js
        {
            "Id": 1,
            "Name": "Harvest Location",
            "LocationTypeId": 1,
            "LocationTypeName": "Default",
            "ForPlantBatches": True,
            "ForPlants": True,
            "ForHarvests": True,
            "ForPackages": True
        }
    ```
    """

    # def __init__(self, client, properties, license_number=''):
    #     super().__init__(client, properties, license_number)
    #     self._parameters = {
    #         'name': 'Name',
    #         'location_type': 'LocationTypeName',
    #         'batches': 'ForPlantBatches',
    #         'plants': 'ForPlants',
    #         'harvests': 'ForHarvests',
    #         'packages': 'ForPackages'
    #     }
    
    def create(self, license_number='', return_obs=False):
        """Create a location in Metrc."""
        self.client.create_location(
            self.name,
            self.location_type,
            license_number,
            return_obs=return_obs,
        )

    def update(self, **kwargs):
        """Update the location."""
        data = self.to_dict()
        update = clean_dictionary(data, camelcase)
        for param in kwargs:
            # key = self._parameters.get(param, param)
            key = LOCATION_FIELDS.get(param, param)
            update[key] = kwargs[param]
        self.client.update_locations([update])

    def delete(self):
        """Delete the location."""
        self.client.delete_location(self.id)


class Harvest(Model):
    """A class that represents a cannabis harvest. Metrc documentation states:

    A harvest batch is created and given a unique Harvest Name when plants
    or plant material are harvested. [The following 7 conditions must be met.]

    1. Harvest Name - Harvests must be strain specific. The Harvest Name must
    be unique. It is a best practice for the harvest name to include the Strain
    Name and Harvest Date, but it is not required by the State.

    2. Weight - The plant is weighed individually in its entirety after being
    cut from root ball (stem, stalk, bud/flower, leaves, trim leaves, etc.).

    3. Waste - This can be recorded using multiple entries but must be reported
    within three days of destruction.

    4. Package - Package and tag the product from the Harvest Batch (Fresh
    Cannabis Plant, Flower, Leaf or Kief). These packages must be strain
    specific.

    5. Transfer - Licensee must create transfer manifest to move product to a
    Processor, Distributor, or Manufacturer.

    6. Finish - When the Harvest Batch (HB) has been fully packaged, there
    should be remaining wet weight to account for moisture loss. Selecting
    Finish Harvest will attribute any remaining weight to moisture loss.

    7. A Harvest Batch package of Flower, Leaf, Kief or Fresh Cannabis Plant
    can only be created from the Harvested Tab using a single strain from
    plants harvested at the same time.
    """

    def create_package(
            self,
            name,
            tag,
            weight,
            location=None,
            note='',
            uom=None
        ):
        """Create a package from a harvest.
        Args:
            name (str): The name of the packaged item.
            tag (str): A tag for the package.
            weight (float): The weight of the package.
            location (str): An optional location for the package, the location
                of the harvest is used by default.
            note (str): A note for the package.
            uom (str): The unit of measure, the unit of measure from the
                harvest is used by default.
        """
        data = {
            'Tag': tag,
            'Location': location or self.drying_location_name,
            'Item': name,
            'UnitOfWeight': uom or self.unit_of_weight_name,
            'Note': note,
            'ActualDate': get_timestamp(zone=self.client.state),
            'Ingredients': [
                {
                    'HarvestId': self.id,
                    'Weight': weight,
                    'UnitOfWeight': uom or self.unit_of_weight_name,
                },
            ]
        }
        self.client.create_harvest_packages(
            [data],
            license_number=self._license,
        )
        # TODO: Implement return_obs

    def create_packages(
            self,
            name,
            tag,
            weights,
            location=None,
            note='',
            uom=None
        ):
        """Create packages from a harvest.
        Args:
            name (str): The name of the packaged item.
            tag (str): A tag for the package.
            weights (list): A list of package weights (float).
            location (str): An optional location for the package, the location
                of the harvest is used by default.
            note (str): A note for the package.
            uom (str): The unit of measure, the unit of measure from the
                harvest is used by default.
        """
        if uom is None:
            uom = self.unit_of_weight_name
        data = {
            'Tag': tag,
            'Location': location or self.drying_location_name,
            'Item': name,
            'UnitOfWeight': uom,
            'Note': note,
            'ActualDate': get_timestamp(zone=self.client.state),
            'Ingredients': []
        }
        for weight in weights:
            data['Ingredients'].append({
                'HarvestId': self.id,
                'Weight': weight,
                'UnitOfWeight': uom,
            })
        self.client.create_harvest_packages(data, license_number=self._license)
        # TODO: Implement return_obs

    def create_testing_packages(self, data):
        """Create testing packages from a harvest.
        Args:
            data (list): The package data.
        """
        self.client.create_harvest_testing_packages(data, license_number=self._license)
        # TODO: Implement return_obs

    def remove_waste(self, weight, waste_type='Waste', uom='Grams'):
        """Remove waste from the harvest.
        Args:
            weight (float): Required harvest weight.
            waste_type (str): The type of waste.
            uom (str): The unit of measure.
        """
        data = {
            'Id': self.uid,
            'WasteType': waste_type,
            'UnitOfWeight': uom,
            'WasteWeight': weight,
            'ActualDate': get_timestamp(zone=self.client.state)
        }
        self.client.remove_waste([data], license_number=self._license)

    def finish(self):
        """Finish a harvest."""
        data = {
            'Id': self.uid,
            'ActualDate': get_timestamp(zone=self.client.state),
        }
        self.client.finish_harvests([data], license_number=self._license)

    def unfinish(self):
        """Un-finish a harvest."""
        self.client.unfinish_harvests(
            [{'Id': self.uid}],
            license_number=self._license,
        )

    def move(self, destination, harvest_name=None):
        """Move a harvest.
        Args:
            destination (str): THe name of the destination location.
            harvest_name (str): An optional harvest name.
        """
        data = {
            "Id": self.uid,
            "HarvestName": harvest_name,
            "DryingLocation": destination,
            "ActualDate": get_timestamp(zone=self.client.state)
        }
        self.client.move_harvest([data], license_number=self._license)


class Package(Model):
    """A class that represents a cannabis package. Metrc documentation states:

    Immature plants and seeds can be packaged by a nursery and transported by a
    distributor to a cultivator, distributor or retailer for sale.

    2. When a manufacturer is creating a concentrate that will then be used in
    multiple infused production batches, the concentrate must be created as a
    new package. The infused production batches will then be created from the
    concentrate package.

    A. The new package of concentrate is a production batch and will then be
    partially used in an infused product or sold to a customer.

    B. This makes it more easily recorded as connected to the finished infused
    product package.

    3. Packages made at a manufacturer facility that creates concentrates must
    be created by pulling from other packages.

    4. A package must exist in order for it to be selected for transfer.
    Transfers are realtime inventory dependent.

    5. There must be a contents section for each new package created from an
    existing package.

    6. When adjusting a package, use the appropriate adjustment reason.

    7. In order for a distributor to send a sample for testing, a test sample
    package must be created. A new test sample must have a new RFID package tag
    and be pulled from an existing package.

    8. Package tags may only be used once and may not be reused.
    """

    def create_package(
            self,
            name,
            tag='',
            label='',
            labels=[],
            weight=0,
            weights=[],
            location=None,
            note='',
            patient_license=None,
            uom=None,
            uoms=[],
            production_batch=False,
            donation=False,
            remediation=False,
            same_item=False,

        ):
        """Create a package from a harvest.
        Args:
            name (str): The name of the packaged item.
            tag (str): A tag for the package.
            label (str): A package label for the item.
            labels (list): A list of package labels (str) that supersede the
                label.
            weight (float): A weight for the package.
            weights (list): A list of weights (float) for the packages that
                supersede the weight.
            location (str): An optional location for the package, the location
                of the harvest is used by default.
            patient_license (str): A patient license number.
            note (str): A note for the package.
            uom (str): The unit of measure for the package and items.
            uoms (list): A list of unit of measures for the items.
            production_batch (bool): Whether the package is for production.
            donation (bool): Whether the package is for donation.
            remediation (bool): Whether the package is for remediation.
            same_item (bool): Whether the package has the same items as its
                parent.
        """
        if weights:
            total_weight = sum(weights)
        else:
            total_weight = weight
        data = {
            'Tag': tag,
            'Location': location or self.location_name,
            'Item': name,
            'Quantity': total_weight,
            'UnitOfMeasure': uom or self.unit_of_measure_name,
            'PatientLicenseNumber': patient_license,
            'Note': note,
            'IsProductionBatch': bool(production_batch),
            'ProductionBatchNumber': production_batch,
            'IsDonation': donation,
            'ProductRequiresRemediation': remediation,
            'UseSameItem': same_item,
            'ActualDate': get_timestamp(zone=self.client.state),
            'Ingredients': []
        }

        if labels:
            for index, item_label in enumerate(labels):
                try:
                    item_uom = uoms[index]
                except IndexError:
                    item_uom = uom or self.unit_of_measure_name
                try:
                    item_weight = weights[index]
                except IndexError:
                    item_weight = weight
                data['Ingredients'].append({
                    'Package': item_label,
                    'Quantity': item_weight,
                    'UnitOfMeasure': item_uom,
                })
        else:
            data['Ingredients'].append({
                'Package': label,
                'Quantity': weight,
                'UnitOfMeasure': uom or self.unit_of_measure_name,
            })
        self.client.create_packages(
            [data],
            license_number=self._license,
        )
        # TODO: Implement return_obs
    

    def create_plant_batch(
            self,
            name,
            count,
            weight,
            location=None,
            batch_type='Clone',
            units='Ounces',
            date=None,
        ):
        """Create an immature plant batch from the package.
        Args:
            name (str): The name of the new plant batch.
            count (int): The number of clones being planted.
            location (str): An optional new location for the plant batch.
            batch_type (str): The type of planting, Seed of Clone, with Clone
                as the default.
        """
        if date is None:
            date = get_timestamp(zone=self.client.state)
        data = {
            'PackageLabel': self.label,
            'PackageAdjustmentAmount': weight,
            'PackageAdjustmentUnitOfMeasureName': units,
            'PlantBatchName': name,
            'PlantBatchType': batch_type,
            'PlantCount': count,
            'LocationName': location or self.location_name,
            'StrainName': self.strain_name,
            'PatientLicenseNumber': None,
            'PlantedDate': get_timestamp(zone=self.client.state),
            'UnpackagedDate': date,
        },
        self.client.manage_packages(
            [data],
            action='create/plantings',
            license_number=self._license
        )

    def change_item(self, item_name):
        """Change the item of the package."""
        data = {
            'Label': self.label,
            'Item': item_name,
        }
        self.client.change_package_items([data], self._license)

    def finish(self):
        """Finish a package."""
        data = {
            'Label': self.label,
            'ActualDate': get_timestamp(zone=self.client.state)
        }
        self.client.manage_packages(
            [data],
            action='finish',
            license_number=self._license
        )

    def unfinish(self):
        """Un-finish a package."""
        self.client.manage_packages(
            [{'Label': self.label}],
            action='unfinish',
            license_number=self._license
        )

    def adjust(
            self,
            weight,
            note='',
            reason='Mandatory State Destruction',
            uom='Grams'
        ):
        """Adjust the package.
        Args:
            weight (float): Required adjustment weight.
            note (str): Required note for certain reasons.
            reason (str): The reason for adjustment.
            uom (str): The unit of measure.
        """
        data = {
            'Label': self.label,
            'Quantity': weight,
            'UnitOfMeasure': uom,
            'AdjustmentReason': reason,
            'AdjustmentDate': get_timestamp(zone=self.client.state),
            'ReasonNote': note
        }
        self.client.manage_packages(
            [data],
            action='adjust',
            license_number=self._license,
        )

    def remediate(self, method, steps):
        """Remediate the package.
        Args:
            method (str): The method used for remediation.
            steps (str): A description of the steps followed to remediate
                the package.
        """
        data = [{
            'PackageLabel': self.label,
            'RemediationMethodName': method,
            'RemediationDate': get_timestamp(zone=self.client.state),
            'RemediationSteps': steps,
        }]
        self.client.manage_packages(
            data,
            action='remediate',
            license_number=self._license,
        )

    def update_note(self, note):
        """Update the package's note.
        Args:
            note (str): The new package note.
        """
        data = [{
            'PackageLabel': self.label,
            'Note': note,
        }]
        self.client.update_package_notes(data)


    def change_location(self, location):
        """Change the package's location.
        Args:
            location (str): The new location for the package.
        """
        data = [{
            'Label': self.label,
            'Location': location,
            'MoveDate': get_timestamp(zone=self.client.state)
        }]
        self.client.change_package_locations(data)


    def update_items(self, name='', names=[]):
        """Update the package's item.
        Args:
            name (str): The name of the package item.
            names (list): A list of item names to update.
        """
        data = []
        if not names:
            data.append({
                'PackageLabel': self.label,
                'Item': name,
            })
        else:
            # Warning: untested
            for index, item_name in enumerate(names):
                item_name = names[index]
                item_label = self.ingredients[index]['PackageLabel']
                data.append({
                    'PackageLabel': item_label,
                    'Item': item_name,
                })
        self.client.change_package_items(data)


class Patient(Model):
    """A class that represents a cannabis patient.
    ```js
    {
        'PatientId': 1,
        'LicenseNumber': '000001',
        'RegistrationDate': '2015-01-08',
        'LicenseEffectiveStartDate': '2014-07-12',
        'LicenseEffectiveEndDate': '2015-07-07',
        'RecommendedPlants': 6,
        'RecommendedSmokableQuantity': 2.0,
        'HasSalesLimitExemption': false,
        'OtherFacilitiesCount': 1
    }
    ```
    """

    def create(self):
        """Create a patient record in Metrc."""
        context = self.to_dict()
        context['actual_date'] = get_timestamp(zone=self.client.state)
        data = clean_nested_dictionary(context, camelcase)
        self.client.create_patients([data], license_number=self._license)

    def update(self, **kwargs):
        """Update the patient given parameters as keyword arguments."""
        context = self.to_dict().copy()
        data = update_dict(context, **kwargs)
        data = remove_dict_nulls(data)
        self.client.update_patients([data], self._license)

    def delete(self):
        """Delete the patient."""
        self.client.delete_patient(self.id, self._license)


class Plant(Model):
    """A class that represents a cannabis plant. Metrc documentation states:

    Plants are tagged at the immature lot growth phase and at the mature /
    flowering growth phase. A UID number is assigned to an immature plant lot
    of up to 100 seeds or immature plants. Required corresponding UID
    number labels will need to be produced by the licensee.
    
    Once the immature lot has been established in Metrc, the death of an
    immature plant(s) must be recorded in Metrc by
    recording the associated waste amount and reducing the total number of the
    immature plants in the lot for each immature plant that was destroyed.

    Plant tags are assigned to individual plants when they are moved to a
    designated canopy area, or when the plant begins flowering.

    A plant can be destroyed anytime during the growth phases.
    Any waste produced by the plant should be recorded prior to the
    destruction.
    2. Any waste created during the immature growth phase must be recorded as
    waste using the Plant Waste function and destroyed.
    3. When immature plants begin to flower, select the Change Growth Phase
    button to record the change and associate the new Plant Tag ID to the
    plant(s).
    4. In Metrc, anytime something is trimmed from a flowering plant during
    growing with the intent to sell it, process it, or perform a partial
    harvest, a Manicure batch must be created.
    """

    def create_planting(self, name, count, location=None, batch_type='Clone'):
        """Create an immature plant batch from the plant.
        Args:
            name (str): The name of the new plant batch.
            count (int): The number of clones being planted.
            location (str): An optional new location for the plant batch.
            batch_type (str): The type of planting, Seed of Clone, with Clone
                as the default.
        """
        data = {
            'PlantLabel': self.label,
            'PlantBatchName': name,
            'PlantBatchType': batch_type,
            'PlantCount': count,
            'LocationName': location or self.location_name,
            'StrainName': self.strain_name,
            'PatientLicenseNumber': None,
            'ActualDate': get_timestamp(zone=self.client.state),
        }
        self.client.manage_plants(
            [data],
            action='create/plantings',
            license_number=self._license
        )
        # TODO: Implement return_obs

    def create_plant_package(
            self,
            name,
            tag,
            count=1,
            batch_type='Clone',
            trade_sample=False,
            donation=False,
            location=None,
            note='',
            patient_license=None,
        ):
        """Create a package of immature plants from the plant.
        Args:
            name (str): The name of the item to create.
            tag (str): The tag for the package.
            count (int): The number of immature plants to package, 1 by
                default.
            batch_type (str): The batch type, Seed or Clone,
                with Clone as the default.
            trade_sample (bool): Whether or not the package is for sale.
            donation (bool): Whether or not the package is for donation.
            location (str): An optional new location for the package.
            note (str): An optional note for the package.
            patient_license (str): An optional patient license.
        """
        data = {
            'PlantLabel': self.label,
            'PackageTag': tag,
            'PlantBatchType': batch_type,
            'Item': name,
            'Location': location or self.location_name,
            'Note': note,
            'IsTradeSample': trade_sample,
            'PatientLicenseNumber': patient_license,
            'IsDonation': donation,
            'Count': count,
            'ActualDate': get_timestamp(zone=self.client.state)
        }
        self.client.manage_plants(
            [data],
            action='create/plantbatch/packages',
            license_number=self._license
        )
        # TODO: Implement return_obs

    def flower(self, tag, location_name=None):
        """Change the growth phase of the plant to flowering.
        Args:
            tag (str): A tag to assign to the flowering plant.
            location_name (str): An optional new location for the plant.
        """
        if location_name is None:
            location_name = self.location_name
        data = {
            'Label': self.label,
            'NewTag': tag,
            'GrowthPhase': 'Flowering',
            'NewLocation': location_name,
            'GrowthDate': get_timestamp(zone=self.client.state)
        }
        self.client.manage_plants(
            [data],
            action='changegrowthphases',
            license_number=self._license
        )

    def move(self, location_name):
        """Move the plant to a new location.
        Args:
            location_name (str): The destination's name.
        """
        data = {
            'Id': self.id,
            'Location': location_name,
            'ActualDate': get_timestamp(zone=self.client.state)
        }
        self.client.manage_plants(
            [data],
            action='moveplants',
            license_number=self._license
        )

    def destroy(
            self,
            weight,
            method='Compost',
            material='Soil',
            note='n/a',
            reason='Contamination',
            uom='grams',
        ):
        """Destroy the plant.
        Args:
            weight (float): Required weight of the waste.
            material (str): The waste material, e.g soil.
            method (str): The mechanism of destruction:
                `Grinder` or `Compost`.
            reason (str): The reason for destruction:
                `Contamination` or `Male Plants`.
        """
        data = {
            'Id': self.id,
            'WasteMethodName': method,
            'WasteMaterialMixed': material,
            'WasteWeight': weight,
            'WasteUnitOfMeasureName': uom,
            'WasteReasonName': reason,
            'ReasonNote': note,
            'ActualDate': get_timestamp(zone=self.client.state)
        }
        self.client.manage_plants(
            [data],
            action='destroyplants',
            license_number=self._license
        )

    def manicure(
            self,
            weight,
            harvest_name=None,
            location_name=None,
            patient_license=None,
            uom='Grams',
        ):
        """Manicure the plant.
        Args:
            weight (float): Required harvest weight.
            harvest_name (str): Optional harvest name.
            location_name (str): The drying location's name.
            patient_license (str): A patient's license number.
            uom (str): The unit of measure
        """
        if location_name is None:
            location_name = self.location_name
        data = {
            'Plant': self.label,
            'Weight': weight,
            'UnitOfWeight': uom,
            'DryingLocation': location_name,
            'HarvestName': harvest_name,
            'PatientLicenseNumber': patient_license,
            'ActualDate': get_timestamp(zone=self.client.state)
        }
        self.client.manage_plants(
            [data],
            action='manicureplants',
            license_number=self._license
        )
        # TODO: Implement return_obs

    def harvest(
            self,
            harvest_name,
            weight,
            location_name=None,
            patient_license=None,
            uom='Grams',
        ):
        """Harvest the plant.
        Args:
            harvest_name (str): Required harvest name.
            weight (float): Required harvest weight.
            location_name (str): The harvest location's name.
            patient_license (str): A patient's license number.
            uom (str): The unit of measure.
        """
        if location_name is None:
            location_name = self.location_name
        data = {
            'Plant': self.label,
            'Weight': weight,
            'UnitOfWeight': uom,
            'DryingLocation': location_name,
            'HarvestName': harvest_name,
            'PatientLicenseNumber': patient_license,
            'ActualDate': get_timestamp(zone=self.client.state)
        }
        self.client.manage_plants(
            [data],
            action='harvestplants',
            license_number=self._license
        )
        # TODO: Implement return_obs


class PlantBatch(Model):
    """A class that represents a cannabis plant batch.
    ```js
        {
            "Id": 5,
            "Name": "Demo Plant Batch 1",
            "Type": "Seed",
            "LocationId": null,
            "LocationName": null,
            "LocationTypeName": null,
            "StrainId": 1,
            "StrainName": "Spring Hill Kush",
            "PatientLicenseNumber": null,
            "UntrackedCount": 80,
            "TrackedCount": 10,
            "PackagedCount": 0,
            "HarvestedCount": 0,
            "DestroyedCount": 40,
            "SourcePackageId": null,
            "SourcePackageLabel": null,
            "SourcePlantId": null,
            "SourcePlantLabel": null,
            "SourcePlantBatchId": null,
            "SourcePlantBatchName": null,
            "PlantedDate": "2014-10-10",
            "LastModified": "0001-01-01T00:00:00+00:00"
        }
        {
            "Name": "B. Kush 5-30",
            "Type": "Clone",
            "Count": 25,
            "Strain": "Spring Hill Kush",
            "Location": null,
            "PatientLicenseNumber": "X00001",
            "ActualDate": "2015-12-15"
        }
    ```
    """

    RETURNED_VALUES = {
        'TrackedCount': 'count',
        'StrainName': 'strain',
        'LocationName': 'location',
        'QuantityType': 'quantity_type',
        'DefaultLabTestingState': 'default_lab_testing_state',
        'ApprovalStatus': 'approval_status',
        'ApprovalStatusDateTime': 'approval_status_date_time',
        'StrainId': 'strain_id',
        'AdministrationMethod': 'administration_method',
        'UnitQuantity': 'unit_quantity',
        'UnitQuantityUnitOfMeasureName': 'unit_quantity_unit_of_measure_name',
    }

    def __init__(self, client, properties, license_number=''):
        super().__init__(client, properties, license_number)
        for k, v in self.RETURNED_VALUES.items():
            try:
                self.__dict__[v] = properties[k]
            except KeyError:
                pass

    def add_additive(self,
            name,
            amount,
            units='Gallons',
            additive_type='Fertilizer',
            device=None,
            supplier=None,
            date=None,
            ingredients=[],
        ):
        """Add an additive to the plant batch..
        Args:
            count (int): The number of plants to destroy.
            ingredients (list): A list of ingredients as dictionaries,
                e.g. {'name': 'Nitrogen', 'percentage': 4.20}
        """
        if date is None:
            date = get_timestamp(zone=self.client.state)
        data = {
            'AdditiveType': additive_type,
            'ProductTradeName': name,
            'EpaRegistrationNumber': None,
            'ProductSupplier': supplier,
            'ApplicationDevice': device,
            'TotalAmountApplied': amount,
            'TotalAmountUnitOfMeasure': units,
            'PlantBatchName': self.name,
            'ActualDate': date,
            'ActiveIngredients': [
                clean_dictionary(x, camelcase)(x) for x in ingredients
            ]
        }
        self.client.manage_batches([data], 'additives', self._license)

    def create(self, license_number=''):
        """Create a plant batch record in Metrc."""
        context = self.to_dict()
        data = clean_dictionary(context, camelcase)
        self.client.manage_batches([data], 'createplantings', license_number)

    def create_package(
            self,
            item_name,
            tag,
            count,
            location='',
            note='',
            trade_sample=False,
            donation=False,
        ):
        """Create a package from the plant batch."""
        data = {
            'id': self.uid,
            'count': count,
            'location': location or self.location_name,
            'item': item_name,
            'tag': tag,
            'note': note,
            'is_trade_sample': trade_sample,
            'is_donation': donation,
            'actual_date': get_timestamp(zone=self.client.state)
        }
        data = clean_dictionary(data, camelcase)
        self.client.manage_batches([data], 'createpackages', self._license)
        # TODO: Implement return_obs

    # TODO: Implement a `create_packages` to create multiple packages.

    def create_package_from_mother(
            self,
            tag,
            item,
            count,
            location=None,
            note='',
            trade_sample=False,
            donation=False,
        ):
        """Create a package from the plant batch mother plant."""
        data = {
            'PlantBatch': self.name,
            'Count': count,
            'Location': location or self.location_name,
            'Item': item,
            'Tag': tag,
            'Note': note,
            'IsTradeSample': trade_sample,
            'IsDonation': donation,
            'ActualDate': get_timestamp(zone=self.client.state),
        }
        data = clean_dictionary(data, camelcase)
        self.client.manage_batches(
            [data],
            '/create/packages/frommotherplant',
            self._license,
        )
        # TODO: Implement return_obs

    def change_growth_phase(
            self,
            tag,
            count=1,
            growth_phase='Vegetative',
            location=None,
            patient_license=None,
        ):
        """Change the growth phase of the batch.
        Args:
            tag (str): A plant tag for the new growth phase for the plants.
                Subsequent tags will be used for plants beyond the first.
            count (int): The number of plants to change growth phase, 1 by
                default.
            growth_phase (str): The growth phase for the plant(s), Flowering
                or Vegetative. Vegetative is used by default.
            location (str): A location for the plants, the current location of
                the plant batch is used by default.
            patient_license (str): An optional patient license number.
        """
        data = {
            'Name': self.name,
            'Count': count,
            'StartingTag': tag,
            'GrowthPhase': growth_phase,
            'NewLocation': location or self.location_name,
            'GrowthDate': get_timestamp(zone=self.client.state),
            'PatientLicenseNumber': patient_license,
        }
        self.client.manage_batches([data], 'changegrowthphase', self._license)

    def destroy_plants(self, count, reason):
        """Destroy a number of plants for a given reason.
        Args:
            count (int): The number of plants to destroy.
            reason (str): The reason for the destruction.
        """
        data = {
            'PlantBatch': self.name,
            'Count': count,
            'ReasonNote': reason,
            'ActualDate': get_timestamp(zone=self.client.state),
        }
        self.client.manage_batches([data], 'destroy', self._license)

    def split(self, name, count, location=None):
        """Split the batch.
        Args:
            name (dict): The name of the new batch.
            count (int): The number of plants to include in the new batch.
            location (str): An optional location for the plant.
        """
        data = {
            'PlantBatch': self.name,
            'GroupName': name,
            'Count': count,
            'Location': location or self.location_name,
            'Strain': self.strain_name,
            'PatientLicenseNumber': None,
            'ActualDate': get_timestamp(zone=self.client.state),
        }
        self.client.manage_batches([data], 'split', self._license)
        # TODO: Implement return_obs


class LabResult(Model):
    """A class that represents a cannabis lab result. Metrc documentation
    states:

    The Lab Results tab displays the details of each individual lab test
    performed on the package. A Document Download button is available on each
    row on the Lab Results tab to view the associated certificate of analysis
    (COA), which the laboratory staff uploads when test results are recorded.
    The test results and COA are available on the source package and any
    related packages only after the laboratory staff releases the results.
    """

    def create(self, data={}):
        """Post lab result data."""
        context = self.to_dict()
        result = clean_dictionary(data, camelcase)
        self.client.post_lab_results([{**context, **result}], self._license)

    def post(self, data={}):
        """Post lab result data. Equivalent alternative of `create`."""
        self.create(data)

    def upload_coa(self, data={}):
        """Upload lab result CoA."""
        context = self.to_dict()
        result = clean_dictionary(data, camelcase)
        self.client.upload_coas([{**context, **result}], self._license)

    def release(self, data={}):
        """Release lab results."""
        context = self.to_dict()
        result = clean_dictionary(data, camelcase)
        self.client.release_lab_results([{**context, **result}], self._license)


class Receipt(Model):
    """A class that represents a cannabis sale receipt. Sales are reported to
    record the transfer of cannabis products to a consumer, patient or
    caregiver.

    When you request receipts you receive the following object.
    ```js
    {
        "Id": 1,
        "ReceiptNumber": null,
        "SalesDateTime": "2016-01-01T17:35:45.000",
        "SalesCustomerType": "Consumer",
        "PatientLicenseNumber": null,
        "CaregiverLicenseNumber": null,
        "IdentificationMethod": null,
        "TotalPackages": 0,
        "TotalPrice": 0.0,
        "Transactions": [],
        "IsFinal": false,
        "ArchivedDate": null,
        "RecordedDateTime": "0001-01-01T00:00:00+00:00",
        "RecordedByUserName": null,
        "LastModified": "0001-01-01T00:00:00+00:00"
    }
    ```

    When you create a receipt, you pass the following object.
    ```js
    {
        "SalesDateTime": "2016-10-04T16:44:53.000",
        "SalesCustomerType": "Consumer",
        "PatientLicenseNumber": null,
        "CaregiverLicenseNumber": null,
        "IdentificationMethod": null,
        "Transactions": [
            {
                "PackageLabel": "ABCDEF012345670000010331",
                "Quantity": 1.0,
                "UnitOfMeasure": "Ounces",
                "TotalAmount": 9.99
            }
        ]
    }
    ```
    """

    RETURNED_VALUES = {
        'ReceiptNumber': 'receipt_number',
        'TotalPackages': 'total_packages',
        'TotalPrice': 'total_price',
        'IsFinal': 'is_final',
        'RecordedDateTime': 'recorded_date_time',
        'RecordedByUserName': 'recorded_by_user_name',
        'LastModified': 'last_modified',
    }

    def create(self):
        """Create a receipt record in Metrc."""
        context = self.to_dict()
        data = clean_nested_dictionary(context, camelcase)
        self.client.create_receipts([data], license_number=self._license)

    def update(self, **kwargs):
        """Update the receipt given parameters as keyword arguments."""
        context = self.to_dict().copy()
        data = update_dict(context, **kwargs)
        data = remove_dict_fields(data, self.RETURNED_VALUES.keys())
        data = remove_dict_nulls(data)
        self.client.update_receipts([data], self._license)

    def delete(self):
        """Delete the receipt."""
        self.client.delete_receipt(self.id, self._license)


class Strain(Model):
    """A class that represents a cannabis strain.
    ```js
        {
            "Id": 1,
            "Name": "Old-time Moonshine",
            "TestingStatus": "InHouse",
            "ThcLevel": 0.1865,
            "CbdLevel": 0.1075,
            "IndicaPercentage": 25.0,
            "SativaPercentage": 75.0
        }
    ```
    """

    def create(self):
        """Create a strain record in Metrc."""
        context = self.to_dict()
        data = clean_dictionary(context, camelcase)
        self.client.create_strains([data])

    def update(self, **kwargs):
        """Update the strain given parameters as keyword arguments."""
        context = self.to_dict()
        data = update_dict(context, camelcase, **kwargs)
        self.client.update_strains([data], license_number=self._license)

    def delete(self):
        """Delete the strain."""
        self.client.delete_strain(self.id, license_number=self._license)


class Transfer(Model):
    """A class that represents a cannabis transfer. Metrc documentation states:

    Transfers are a key component of the chain of custody process.
    A transfer must be created anytime a package moves from one licensee to
    another, even if the two facilities are located on the same property.

    Packages can only be transported from one licensee to another by a licensed
    Distributor. A Testing Laboratory is allowed to transport test samples for
    official state testing. Distributors and Testing Laboratories are required
    to record the actual departure time from the origin facility and the actual
    arrival time at the destination facility in Metrc real-time.

    A package must be received in its entirety (the system DOES NOT allow
    receiving a partial package). A transfer can be rejected by individual
    package, or in whole by rejecting all packages. A rejected package requires
    the originating Licensee to receive the package back into inventory.
    A package must exist in order to be selected for transfer. Transfers are
    done in real time and are inventory dependent. When receiving a package,
    any adjustments to the weight, volume, or count may be reported to the
    State. If there are any questions about a transfer, reject it.

    A transfer can be modified , or voided, up until the time that the
    Distributor or Testing Laboratory marks that it has departed the facility.
    Once the transfer process has begun, the transfer may not be modified
    except by the Distributor or Testing Laboratory to edit estimated departure
    and arrival times, or driver and vehicle information(see Edit Transporter
    Info below).

    When modifying transfers, each of the transfer fields may be modified at
    the same level of detail as when the transfer was created. Edits may be
    completed for a variety of reasons including: error correction, changes in
    destination, changes in product, etc.

    Voiding a transfer can only be completed by the originating business.
    Voiding a transfer permanently eliminates it and moves the product back
    into the originator’s inventory. Once a transfer has been voided, it cannot
    be reinstated and all associated packages will be returned to the transfer
    originator's inventory.

    Receiving a transfer is the final point of exchange in the chain of
    custody.

    Type `Transfer` is used for all transfers except transfers requiring the
    use of a `Wholesale Manifest` or `Return` manifest. A `Return` transfer is
    used only for the transfer of defective manufactured products back to the
    originating licensee. A `Wholesale Manifest` transfer is used when =
    transferring products to a Retailer licensee. When a `Wholesale Manifest`
    is used, the originator is required to record the wholesale price of each
    package in the transfer. A Microbusiness functioning as a Distributor with
    a transfer that includes a Retailer licensee, or another Microbusiness
    licensee functioning as a Retailer, shall follow the process above. It is
    recommended that Nurseries utilize a `Wholesale Manifest` when transferring
    seeds or immature plants to a Retailer.
    """

    RETURNED_VALUES = {}

    def create(self):
        """Create a transfer record in Metrc."""
        context = self.to_dict()
        data = clean_nested_dictionary(context, camelcase)
        self.client.create_transfers([data], license_number=self._license)

    def update(self, **kwargs):
        """Update the transfer given parameters as keyword arguments."""
        context = self.to_dict().copy()
        data = update_dict(context, **kwargs)
        data = remove_dict_fields(data, self.RETURNED_VALUES.keys())
        data = remove_dict_nulls(data)
        self.client.update_transfers([data], self._license)

    def delete(self):
        """Delete the transfer."""
        self.client.delete_transfer(self.id, self._license)


class TransferTemplate(Model):
    """A class that represents a cannabis transfer template. The template can
    be copied to create other templates. Transfer templates can be used for
    transfers to the same destination licensee utilizing the
    same:
        - Planned Route
        - Transporter(s)
        - Driver(s)
        - Vehicle(s)
        - Package
    """

    RETURNED_VALUES = {}

    def create(self):
        """Create a transfer template record in Metrc."""
        context = self.to_dict()
        data = clean_nested_dictionary(context, camelcase)
        self.client.create_transfer_templates(
            [data],
            license_number=self._license,
        )

    def update(self, **kwargs):
        """Update the transfer template given parameters as keyword
        arguments."""
        context = self.to_dict().copy()
        data = update_dict(context, **kwargs)
        data = remove_dict_fields(data, self.RETURNED_VALUES.keys())
        data = remove_dict_nulls(data)
        self.client.update_transfer_templates([data], self._license)

    def delete(self):
        """Delete the transfer template."""
        self.client.delete_transfer_template(self.id, self._license)


class Transaction(Model):
    """A class that represents a cannabis sale transaction. When you get a
    transaction you receive an object as follows.
    ```js
    {
        "SalesDate": "2015-01-08",
        "TotalTransactions": 40,
        "TotalPackages": 40,
        "TotalPrice": 399.6
    }
    ```
    A created transaction is as follows.
    ```js
    {
        "PackageId": 71,
        "PackageLabel": "ABCDEF012345670000010331",
        "ProductName": "Shake",
        "ProductCategoryName": null,
        "ItemStrainName": null,
        "ItemUnitCbdPercent": null,
        "ItemUnitCbdContent": null,
        "ItemUnitCbdContentUnitOfMeasureName": null,
        "ItemUnitCbdContentDose": null,
        "ItemUnitCbdContentDoseUnitOfMeasureName": null,
        "ItemUnitThcPercent": null,
        "ItemUnitThcContent": null,
        "ItemUnitThcContentUnitOfMeasureName": null,
        "ItemUnitThcContentDose": null,
        "ItemUnitThcContentDoseUnitOfMeasureName": null,
        "ItemUnitVolume": null,
        "ItemUnitVolumeUnitOfMeasureName": null,
        "ItemUnitWeight": null,
        "ItemUnitWeightUnitOfMeasureName": null,
        "ItemServingSize": null,
        "ItemSupplyDurationDays": null,
        "ItemUnitQuantity": null,
        "ItemUnitQuantityUnitOfMeasureName": null,
        "QuantitySold": 1.0,
        "UnitOfMeasureName": "Ounces",
        "UnitOfMeasureAbbreviation": "oz",
        "TotalPrice": 9.99,
        "SalesDeliveryState": null,
        "ArchivedDate": null,
        "RecordedDateTime": "0001-01-01T00:00:00+00:00",
        "RecordedByUserName": null,
        "LastModified": "0001-01-01T00:00:00+00:00"
    }
    ```
    When you update a transaction, you pass the following object.
    ```js
    {
        "PackageLabel": "ABCDEF012345670000010331",
        "Quantity": 1.0,
        "UnitOfMeasure": "Ounces",
        "TotalAmount": 9.99
    }
    ```
    """

    RETURNED_VALUES = {}

    def create(self):
        """Create a transaction record in Metrc."""
        context = self.to_dict()
        data = clean_nested_dictionary(context, camelcase)
        self.client.create_transactions([data], license_number=self._license)

    def update(self, **kwargs):
        """Update the transaction given parameters as keyword arguments."""
        context = self.to_dict().copy()
        data = update_dict(context, **kwargs)
        data = remove_dict_fields(data, self.RETURNED_VALUES.keys())
        data = remove_dict_nulls(data)
        self.client.update_transactions([data], self._license)


# TODO: Create a Job or ProcessingJob class.


class Waste(Model):
    """A class that represents cannabis waste. Metrc documentation states:

    A harvest batch is created and given a unique Harvest Name when plants or
    plant material are harvested. [The following regulations need to be met.]

    1. Plant waste must be recorded within three business days of destruction.
    In Metrc plant waste can be recorded by Immature Plant Lot, Flowering Plant
    or by Location. Waste can also be recorded by Harvest Batch. See the Metrc
    User Guide for details.

    2. When recording Flowering Plant waste, the waste from multiple plants can
    be recorded as a single waste event but the flowering plants contributing
    to the waste must be individually identified.

    3. If a plant is no longer viable, the waste must be recorded prior to
    recording its destruction.

    4. The reason for the waste must be identified using the Waste Reasons
    defined by the State of California as listed in Exhibit 38 below. Use of
    some Waste Reasons may be limited to certain license types, as determined
    by the State.
    """
    pass



"""
Invoice Model | Cannlytics
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/5/2021
Updated: 11/5/2021
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description: Invoice data model.
"""
# Standard imports.
from dataclasses import dataclass, field
from typing import List

# Internal imports.
from .base import Model


@dataclass
class Invoice(Model):
    """Invoices are incoming or outgoing bills that you want to manage
    in the Cannlytics platform."""
    _collection = 'organizations/%s/invoices'
    analyses: List = field(default_factory=list)



"""
Area Model | Cannlytics
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/5/2021
Updated: 11/5/2021
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description: Area data model.
"""
# Standard imports.
from dataclasses import dataclass
from typing import Optional

# Internal imports.
from .base import Model


@dataclass
class Area(Model):
    """An abstract area that your organization uses in your workflow.
    Areas represent distinct places at your facility(ies).
    Areas are abstract units where you can store inventory,
    plants, samples, instruments, or whatever you please."""
    _collection = 'organizations/%s/areas'
    active: bool = True
    area_type: str = 'General'
    area_type_id: str = ''
    external_id: str = ''
    name: str = ''
    quarantine: bool = False
    for_batches: Optional[bool] = False
    for_plants: Optional[bool] = False
    for_harvests: Optional[bool] = False
    for_packages: Optional[bool] = False



"""
Results Model | Cannlytics
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/5/2021
Updated: 11/5/2021
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description: Results data model.
"""
# Standard imports.
from dataclasses import dataclass
from datetime import datetime

# Internal imports.
from .base import Model

@dataclass
class Results(Model):
    """The final result for an analyte of an analysis after the
    appropriate calculation has been applied to the analyte's
    measurement."""
    _collection = 'organizations/%s/results'
    analysis: str = ''
    analysis_status: str = ''
    package_id: str = ''
    package_label: str = ''
    product_name: str = ''
    sample_id: str = ''
    sample_type: str = ''
    tested_at: datetime = None
    voided_at: datetime = None
    released: bool = False
    released_at: datetime = None
    status: str = ''
    result: float = 0.0
    units: str = ''
    notes: str = ''
    non_mandatory: bool = False



"""
StrainsAI
Copyright (c) 2023 Cannlytics

Authors:
    Keegan Skeate <https://github.com/keeganskeate>
Created: 6/29/2023
Updated: 7/2/2023
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>
"""
# Standard imports:
import csv
import json
from time import sleep
from typing import List, Optional

# External imports:
from cannlytics.ai import INSTRUCTIONAL_PROMPT, initialize_openai
import pandas as pd
import openai


def generate_strain_art():
    """Generate a description for a strain."""
    # TODO: Implement!
    raise NotImplementedError


def generate_strain_description():
    """Generate a description for a strain."""
    

    # TODO: Supplement with lab result data!


    raise NotImplementedError


def identify_strains(
        text,
        model='gpt-4',
        max_tokens: Optional[int] = 1_000,
        temperature: Optional[float] = 0.0,
        user: Optional[str] = 'cannlytics',
        verbose: Optional[bool] = False,
        retry_pause: Optional[float] = 3.33,
        instructional_prompt: Optional[str] = None,
        identification_prompt: Optional[str] = None,
        json_key: Optional[str] = 'strains',
    ) -> List[dict]:
    """Identify the closest strain(s) from a strain name."""

    # Format the prompt.
    if instructional_prompt is None:
        instructional_prompt = INSTRUCTIONAL_PROMPT
    if identification_prompt is None:
        identification_prompt = 'Given the following cannabis product name or text, what is the strain(s) of cannabis? Return your answer as JSON, e.g. {"strains": ["X"]} or {"strains": ["Strain 1", "Strain 2"]} where strains is a list of the strains, where sometimes multiple strains are indicated by a cross, e.g. "Strain 1 x Strain 2".'

    # Format the message.
    messages = [
        {'role': 'system', 'content': INSTRUCTIONAL_PROMPT},
        {'role': 'system', 'content': identification_prompt},
        {'role': 'user', 'content': f'Text: {text}\n\nStrains:'}
    ]

    # Make the request to OpenAI. 
    try:
        initialize_openai(openai_api_key)
        response = openai.ChatCompletion.create(
            model=model,
            messages=messages,
            max_tokens=max_tokens,
            temperature=temperature,
            user=user,
        )
    except:
        if retry_pause:
            sleep(retry_pause)
            initialize_openai(openai_api_key)
            response = openai.ChatCompletion.create(
                model=model,
                messages=messages,
                max_tokens=max_tokens,
                temperature=temperature,
                user=user,
            )
    
    # Get the content of the response.
    if verbose:
        print('RESPONSE:', response)
    content = response['choices'][0]['message']['content']
    start_index = content.find('{')
    end_index = content.rfind('}') + 1

    extracted_data = []
    try:
        obj = json.loads(content[start_index:end_index])
        extracted_data.extend(obj[json_key])
    except:
        if verbose:
            print('Failed to extract data.')

    # Return the extracted data.
    # TODO: Return the prompts and cost.
    return extracted_data


def train_strain_name_identification_model():
    """Train a model to identify strain names."""

    # TODO: Get strain name vocabulary.
    strain_vocab_file = 'strain-vocab.txt'
    strain_data = pd.read_excel('ccrs-strain-statistics-2023-03-07.xlsx')
    strain_data['strain_name'] = strain_data['strain_name'].apply(
        lambda x: x.title()
    )
    strain_data.drop_duplicates(subset=['strain_name'], inplace=True)
    strain_data.sort_values(by=['strain_name'], inplace=True)
    strain_data.to_csv(
        strain_vocab_file,
        header=None,
        index=None,
        sep='\t',
        mode='w',
        columns=['strain_name'],
        quoting=csv.QUOTE_NONE,
        # escapechar=None
    )

    # TODO: Fine-tine a model to identify strain names.

    raise NotImplementedError


# === Tests ===
# Performed 2023-07-01 by Keegan Skeate <admin@cannlytics.com>.
if __name__ == '__main__':

    from dotenv import dotenv_values
    print('Testing strains AI...')

    # Initialize OpenAI.
    config = dotenv_values('../../../.env')
    openai_api_key = config['OPENAI_API_KEY']

    # [✓] TEST: Identify strain names.
    text = 'GARCIA HAND PICKED DARK KARMA'
    strain_names = identify_strains(text)
    print(strain_names)



"""
Parse TagLeaf LIMS CoA
Copyright (c) 2022-2023 Cannlytics

Authors:
    Keegan Skeate <https://github.com/keeganskeate>
    Candace O'Sullivan-Sutherland <https://github.com/candy-o>
Created: 7/15/2022
Updated: 3/21/2023
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description:

    Parse a TagLeaf LIMS CoA PDF or URL. Labs include:

        - 2 River Labs, Inc
        - BelCosta Labs

Data Points:

    ✓ analyses
    ✓ {analysis}_method
    ✓ {analysis}_status
    - coa_urls
    ✓ date_tested
    ✓ date_received
    ✓ distributor
    ✓ distributor_license_number
    ✓ distributor_license_type
    - distributor_latitude (augmented)
    - distributor_longitude (augmented)
    ✓ images
    ✓ lab_results_url
    ✓ producer
    - producer_latitude (augmented)
    - producer_longitude (augmented)
    ✓ product_name
    ✓ product_type
    ✓ results
    - sample_weight
    ✓ status
    ✓ total_cannabinoids
    ✓ total_thc
    ✓ total_cbd
    ✓ total_terpenes (calculated)
    ✓ sample_id (generated)
    - lab_id
    ✓ lab
    ✓ lab_image_url
    ✓ lab_license_number
    ✓ lab_address
    ✓ lab_city (augmented)
    ✓ lab_county (augmented)
    ✓ lab_state (augmented)
    ✓ lab_zipcode (augmented)
    ✓ lab_latitude (augmented)
    ✓ lab_longitude (augmented)
    ✓ lab_phone
    - lab_email

"""
# Standard imports.
from datetime import datetime
import json
import os
from typing import Any, Optional

# External imports.
from bs4 import BeautifulSoup, NavigableString
import pandas as pd
import pdfplumber
from requests import Session

# Internal imports.
from cannlytics import __version__
from cannlytics.data.data import create_hash, create_sample_id
from cannlytics.data.gis import search_for_address
from cannlytics.utils.constants import ANALYSES, STANDARD_UNITS
from cannlytics.utils.utils import (
    convert_to_numeric,
    snake_case,
    strip_whitespace,
)

# It is assumed that the lab has the following details.
TAGLEAF = {
    'coa_algorithm': 'tagleaf.py',
    'coa_algorithm_entry_point': 'parse_tagleaf_coa',
    'lims': 'lims.tagleaf',
    'url': 'https://lims.tagleaf.com',
    'public': True,
}


def parse_tagleaf_url(
        parser,
        url: str,
        headers: Optional[dict] = None,
        keys: Optional[dict] = None,
        persist: Optional[bool] = False,
        **kwargs,
    ) -> dict:
    """Parse a TagLeaf LIMS CoA URL.
    Args:
        url (str): The CoA URL.
        headers (dict): Headers for HTTP requests.
        keys (dict): A dictionary of keys for standardization.
        max_delay (float): Unused argument for standardization.
        persist (bool): Whether to persist the session.
            The default is `False`. If you do persist
            the driver, then make sure to call `quit`
            when you are finished.
    Returns:
        (dict): The sample data.
    """

    # Get the HTML.
    if keys is None:
        keys = parser.fields
    if headers is None:
        headers = parser.headers
    if parser.session is None:
        parser.session = Session()
    response = parser.session.get(url, headers=headers)
    soup = BeautifulSoup(response.content, 'html.parser')

    # Get the date tested.
    obs = {}
    el = soup.find('p', attrs={'class': 'produced-statement'})
    date_tested = pd.to_datetime(el.text.split(': ')[-1]).isoformat()
    obs['date_tested'] = date_tested

    # Get lab details.
    el = soup.find('section', attrs={'class': 'header-container'})
    img = el.find('img')
    pars = el.find_all('p')
    details = [strip_whitespace(x) for x in pars[0].text.split('//')]
    address = details[1]
    obs['lab'] = details[0]
    obs['lab_address'] = address
    obs['lab_image_url'] = img.attrs['src']
    obs['lab_phone'] = details[2].replace('PH: ', '')

    # Try to get lab address details.
    try:
        google_maps_api_key = os.environ['GOOGLE_MAPS_API_KEY']
        location = search_for_address(address, google_maps_api_key)
        obs['lab_street'] = location['street']
        obs['lab_city'] = location['city']
        obs['lab_county'] = location['county']
        obs['lab_state'] = location['state']
        obs['lab_zipcode'] = location['zipcode']
        obs['lab_latitude'] = location['latitude']
        obs['lab_longitude'] = location['longitude']
    except:
        pass

    # Get data from headings.
    text = soup.find_all('p', attrs={'class': 'h5'}, limit=2)
    parts = strip_whitespace(text[0].text.split('//')[0]).split(' (')
    product_name = parts[0]
    obs['product_name'] = product_name
    obs['product_type'] = parts[1].replace(')', '')
    obs['status'] = strip_whitespace(text[1].text.split(':')[-1]).lower()

    # Get cannabinoid totals.
    el = soup.find('div', attrs={'class': 'cannabinoid-overview'})
    rows = el.find_all('div', attrs={'class': 'row'})
    for row in rows:
        pars = row.find_all('p')
        key = snake_case(strip_whitespace(pars[1].text))
        value = strip_whitespace(pars[0].text).replace('%', '').strip()
        if key == 'sample_weight':
            continue
        if key in ['value', 'mg_g', 'lod', 'loq']:
            value = convert_to_numeric(value, strip=True)
        obs[key] = convert_to_numeric(value)

    # Get cultivator and distributor details.
    try:
        els = soup.find_all('div', attrs={'class': 'license'})
        values = [x.text for x in els[0].find_all('p')]
        producer = values[1]
        obs['producer'] = producer
        obs['license_number'] = values[3]
        obs['license_type'] = values[5]
        values = [x.text for x in els[1].find_all('p')]
        obs['distributor'] = values[1]
        obs['distributor_license_number'] = values[3]
        obs['distributor_license_type'] = values[5]
    except:
        client_span = soup.find('span', string='Client')
        producer = client_span.next_sibling.replace(':', '').strip()
        obs['producer'] = producer

    # Get the sample image.
    el = soup.find('div', attrs={'class': 'sample-photo'})
    img = el.find('img')
    image_url = img['src']
    filename = image_url.split('/')[-1]
    obs['images'] = [{'url': image_url, 'filename': filename}]

    # Get the sample details
    el = soup.find('div', attrs={'class': 'sample-info'})
    pars = el.find_all('p')
    for par in pars:
        key = snake_case(par.find('span').text)
        key = keys.get(key, key) # Get preferred key.
        text = par.contents
        value = ''.join([x for x in text if type(x) == NavigableString])
        value = strip_whitespace(value)
        obs[key] = value

    # Get the lab ID and metrc ID.
    obs['lab_id'] = obs.get('sample_id', '')
    metrc_uid = obs.get('source_metrc_uid')
    if metrc_uid:
        obs['metrc_ids'] = [metrc_uid]

    # Format `date_collected` and `date_received` dates.
    try:
        obs['date_collected'] = pd.to_datetime(obs['date_collected']).isoformat()
        obs['date_received'] = pd.to_datetime(obs['date_received']).isoformat()
    except KeyError:
        pass

    # Get the analyses and `{analysis}_status`.
    analyses = []
    el = soup.find('div', attrs={'class': 'tests-overview'})
    blocks = strip_whitespace(el.text)
    blocks = [x for x in blocks.split('    ') if x]
    for i, value in enumerate(blocks):
        if i % 2:
            analysis = analyses[-1]
            if value != '\xa0':
                key = analysis
                if not analysis.endswith('_status'):
                    key = f'{analysis}_status'
                obs[key] = value.lower()
        else:
            analysis = snake_case(value).replace('_status', '')
            analysis = keys.get(analysis, analysis) # Get preferred key.
            analyses.append(analysis)

    # Get `{analysis}_method`s.
    els = soup.find_all('div', attrs={'class': 'table-header'})
    for el in els:
        analysis = el.attrs['id'].replace('_test', '')
        analysis = keys.get(analysis, analysis) # Get preferred key.
        title = el.find('h3').contents
        text = ''.join([x for x in title if type(x) == NavigableString])
        obs[f'{analysis}_method'] = strip_whitespace(text)

    # Get the `results`, using the table header for the columns,
    # noting that `value` is repeated for `mg_g`.
    results = []
    tables = soup.find_all('table')
    for table in tables:

        # Find the analysis of the table.
        title = table.find_previous('h3').contents
        text = ''.join([x for x in title if type(x) == NavigableString])
        analysis = strip_whitespace(text).split(':')[-1].split('by')[0].strip()
        analysis = ANALYSES.get(analysis, snake_case(analysis))
        units = STANDARD_UNITS.get(analysis)

        # Find the columns of the table.
        headers = table.find_all('th')
        columns = [keys[strip_whitespace(x.text)] for x in headers]

        # Iterate over all of the rows of the table.
        rows = table.find_all('tr')[1:]
        for row in rows:
            analyte = None
            mg_g = False
            result = {
                'analysis': analysis,
                'units': units,
            }
            cells = row.find_all('td')
            for i, cell in enumerate(cells):
                key = columns[i]
                value = strip_whitespace(cell.text)
                if key == 'name':
                    analyte = parser.analytes.get(value, snake_case(value))
                    result['key'] = analyte
                if key == 'value' and mg_g:
                    key = 'mg_g'
                if key == 'value':
                    mg_g = True
                if key in ['value', 'mg_g', 'lod', 'loq', 'limit'] and value != 'ND':
                    value = value.replace('< 1 mg/g', 'ND')
                    value = convert_to_numeric(value, strip=True)
                result[key] = value

            # Hot-fix: Handle status.
            if result.get('status') == 'N/A':
                result['status'] = None

            # Handle totals.
            if analyte.startswith('total') and obs.get(analyte) is None:
                obs[analyte] = convert_to_numeric(value, strip=True)
            else:
                results.append(result)
    
    # Hot-fix: Calculate total terpenes.
    terp_results = [x for x in results if 'terp' in x['analysis']]
    if terp_results:
        total_terpenes = 0
        for result in terp_results:
            try:
                total_terpenes += float(result['value'])
            except ValueError:
                pass
        obs['total_terpenes'] = round(total_terpenes, 5)

    # Return the sample with a freshly minted sample ID.
    obs = {**TAGLEAF, **obs}
    obs['lab_results_url'] = url
    obs['analyses'] = json.dumps(list(set(analyses)))
    obs['results'] = json.dumps(results)
    obs['coa_algorithm_version'] = __version__
    obs['coa_parsed_at'] = datetime.now().isoformat()
    obs['results_hash'] = create_hash(results)
    obs['sample_id'] = create_sample_id(
        private_key=json.dumps(results),
        public_key=product_name,
        salt=producer,
    )
    obs['sample_hash'] = create_hash(obs)
    if not persist:
        parser.quit()
    return obs


def parse_tagleaf_pdf(
        parser,
        doc: Any,
        **kwargs,
    ) -> dict:
    """Parse a TagLeaf LIMS CoA PDF.
    Args:
        doc (str or PDF): A PDF file path or pdfplumber PDF.
    Returns:
        (dict): The sample data.
    """
    url = parser.find_pdf_qr_code_url(doc)
    return parse_tagleaf_url(parser, url, **kwargs)


def parse_tagleaf_coa(
        parser,
        doc: Any,
        **kwargs,
    ) -> dict:
    """Parse a TagLeaf LIMS CoA PDF or URL.
    Args:
        doc (str or PDF): A PDF file path or pdfplumber PDF.
    Returns:
        (dict): The sample data.
    """
    if isinstance(doc, str):
        if doc.startswith('http'):
            return parse_tagleaf_url(parser, doc, **kwargs)
        elif doc.endswith('.pdf'):
            data = parse_tagleaf_pdf(parser, doc, **kwargs)
        else:
            data = parse_tagleaf_pdf(parser, doc, **kwargs)
    else:
        data = parse_tagleaf_pdf(parser, doc, **kwargs)
    if isinstance(doc, str):
        data['coa_pdf'] = doc.replace('\\', '/').split('/')[-1]
    elif isinstance(doc, pdfplumber.pdf.PDF):
        data['coa_pdf'] = doc.stream.name.replace('\\', '/').split('/')[-1]
    return data


# === Tests ===
# Uncomment tests to perform them.
# Checked tests were successfully performed by Cannlytics on 3/21/2023.
# Contact: <admin@cannlytics.com>.
if __name__ == '__main__':

    # Test TagLeaf LIMS COA parsing.
    from cannlytics.data.coas import CoADoc
    from dotenv import dotenv_values

    # Set a Google Maps API key.
    config = dotenv_values('../../../.env')
    os.environ['GOOGLE_MAPS_API_KEY'] = config['GOOGLE_MAPS_API_KEY']

    # Specify where your test data lives.
    DATA_DIR = '../../../tests/assets/coas/tagleaf'
    tagleaf_coa_url = 'https://lims.tagleaf.com/coas/F6LHqs9rk9vsvuILcNuH6je4VWCiFzdhgWlV7kAEanIP24qlHS'
    tagleaf_coa_short_url = 'https://lims.tagleaf.com/coa_/F6LHqs9rk9'

    # [✓] TEST: Parse a COA URL.
    # parser = CoADoc()
    # data = parse_tagleaf_url(parser, tagleaf_coa_url)
    # assert data is not None
    # print(data)

    # [✓] TEST: Parse a TagLeaf LIMS COA PDF.
    # parser = CoADoc()
    # tagleaf_coa_pdf = f'{DATA_DIR}/Sunbeam.pdf'
    # data = parse_tagleaf_pdf(parser, tagleaf_coa_pdf)
    # assert data is not None
    # print(data)

    # [✓] TEST: Parse a TagLeaf LIMS COA PDF.
    # parser = CoADoc()
    # tagleaf_coa_pdf = f'{DATA_DIR}/BCL-211214-151.pdf'
    # data = parse_tagleaf_pdf(parser, tagleaf_coa_pdf)
    # assert data is not None
    # print(data)



"""
Instrument Model | Cannlytics
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/5/2021
Updated: 11/5/2021
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description: Instrument data model.
"""
# Standard imports.
from dataclasses import dataclass
from datetime import datetime

# Internal imports.
from .base import Model


@dataclass
class Instrument(Model):
    """A scientific instrument that produces measurements. Instrument
    data is collected through data files, processed with routine
    automation or data importing."""
    _collection = 'instruments/%s'
    area_id: str = ''
    area_name: str = ''
    name: str = ''
    calibrated_at: datetime = None
    calibrated_by: str = ''
    data_path: str = ''
    description: str = ''
    notes: str = ''

    def create_maintenance_log(self):
        """Create a maintenance log."""
        return NotImplementedError



"""
Cannlytics AI
Copyright (c) 2023 Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 6/15/2023
Updated: 7/4/2023
"""
# Standard imports:
import json
import math
import os
from time import sleep
from typing import List, Optional

# External imports:
import google.auth
import openai

# Internal imports:
from cannlytics.firebase import access_secret_version


#-----------------------------------------------------------------------
# OpenAI parameters
#-----------------------------------------------------------------------

# AI warning.
AI_WARNING = "This data was parsed from text using OpenAI's GPT models. Please verify the data before using it. You can submit feedback and report issues to dev@cannlytics.com, thank you."

# OpenAI API model prices (as of 2023-06-06) per 1000 tokens.
PRICE_PER_1000_TOKENS = {
    'gpt-4': {'prompt': 0.03, 'completion': 0.06},
    'gpt-3.5-turbo': {'prompt': 0.002, 'completion': 0.002},
    'ada': {'prompt': 0.0004, 'completion': 0.0004, 'training': 0.0004, 'usage': 0.0016},
    'babbage': {'prompt': 0.0005, 'completion': 0.0005, 'training': 0.0006, 'usage': 0.0024},
    'curie': {'prompt': 0.002, 'completion': 0.002, 'training': 0.003, 'usage': 0.012},
    'davinci': {'prompt': 0.02, 'completion': 0.02, 'training': 0.03, 'usage': 0.12},
    'dalle_1024': {'usage': 0.02},
    'dalle_512': {'usage': 0.018},
    'dalle_256': {'usage': 0.016},
    'whisper': {'usage': 0.006},
}

# Define the maximum number of tokens per prompt.
MAX_PROMPT_LENGTH = 4_000

# Instructional prompt.
INSTRUCTIONAL_PROMPT = 'Only return JSON and always return at least an empty object, {}, if no data can be found. Return a value of `null` for any field that cannot be found.'


#-----------------------------------------------------------------------
# OpenAI initialization.
#-----------------------------------------------------------------------

def initialize_openai(openai_api_key = None) -> None:
    """Initialize OpenAI."""
    if openai_api_key is None:
        try:
            _, project_id = google.auth.default()
            openai_api_key = access_secret_version(
                project_id=project_id,
                secret_id='OPENAI_API_KEY',
                version_id='latest',
            )
        except:
            openai_api_key = os.environ['OPENAI_API_KEY']
    openai.api_key = openai_api_key


#-----------------------------------------------------------------------
# Token and cost management.
#-----------------------------------------------------------------------

def estimate_tokens_of_messages(messages: list, model: Optional[str] = 'gpt-4'):
    """Returns the number of tokens used by a list of messages."""
    tokens_per_message = 4 # every message follows <|start|>{role/name}\n{content}<|end|>\n
    tokens_per_name = 1
    num_tokens = 0
    for message in messages:
        num_tokens += tokens_per_message
        for key, value in message.items():
            num_tokens += math.ceil(len(value) / 4)
            if key == 'name':
                num_tokens += tokens_per_name
    num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
    return num_tokens


def estimate_tokens_of_string(string: str, model: Optional[str] = 'gpt-4') -> int:
    """Returns the number of tokens in a text string."""
    return estimate_tokens_of_messages([{'role': 'user', 'content': string},], model=model)


def get_prompt_price(prompt, model='gpt-4', prices=PRICE_PER_1000_TOKENS):
    """Returns the price to generate a prompt."""
    num_tokens = estimate_tokens_of_string(prompt, model)
    return num_tokens / 1_000 * prices[model]['prompt']


def get_messages_price(messages, model='gpt-4', prices=PRICE_PER_1000_TOKENS):
    """Returns the price to generate a list of messages."""
    num_tokens = estimate_tokens_of_messages(messages, model)
    return num_tokens / 1_000 * prices[model]['prompt']


def get_tokens_price(num_tokens, model='gpt-4', prices=PRICE_PER_1000_TOKENS):
    """Returns the price to generate a number of tokens."""
    return num_tokens / 1_000 * prices[model]['prompt']


def split_string(string, max_length):
    """Split a string into chunks of a given length."""
    return [string[i:i+max_length] for i in range(0, len(string), max_length)]


def split_into_token_chunks(
        text: str,
        max_prompt_length: int,
        model: Optional[str] = 'gpt-4',
    ) -> List[str]:
    """Split a body of text into desired portions less than a given
    desired length in tokens.
    """
    lines = text.split('\n')
    chunks = []
    current_chunk = ''
    for line in lines:
        if estimate_tokens_of_string(current_chunk + '\n' + line, model) <= max_prompt_length:
            current_chunk += '\n' + line
        else:
            chunks.append(current_chunk)
            current_chunk = line
    chunks.append(current_chunk)
    return chunks


#-----------------------------------------------------------------------
# Engineered prompts.
#-----------------------------------------------------------------------

def gpt_to_json(
        text: str,
        system_prompts: List[str],
        model: Optional[str] = 'gpt-4',
        max_tokens: Optional[int] = 4_000,
        temperature: Optional[float] = 0.0,
        user: Optional[str] = 'cannlytics',
        openai_api_key: Optional[str] = None,
        verbose: Optional[bool] = False,
        retry_pause: Optional[float] = 3.33,
        text_label: Optional[str] = 'Text',
        json_label: Optional[str] = 'JSON',
    ) -> dict:
    """Prompt GPT and convert the output content to JSON."""

    # Add the system prompts.
    messages = []
    for system_prompt in system_prompts:
        messages.append({'role': 'system', 'content': system_prompt})
    
    # Add the main prompt.
    main_prompt = f'{text_label}: ' + text + f'\n\n{json_label}:'
    messages.append({'role': 'user', 'content': main_prompt})

    # Make the request to the OpenAI API.
    try:
        if verbose:
            print('MESSAGES:', messages)
        try:
            initialize_openai(openai_api_key)
            response = openai.ChatCompletion.create(
                model=model,
                messages=messages,
                max_tokens=max_tokens,
                temperature=temperature,
                user=user,
            )

        # Retry the request if it fails.
        except:
            if retry_pause:
                if verbose:
                    print('First OpenAI query failed, retrying.')
                sleep(retry_pause)
                initialize_openai(openai_api_key)
                response = openai.ChatCompletion.create(
                    model=model,
                    messages=messages,
                    max_tokens=max_tokens,
                    temperature=temperature,
                    user=user,
                )
        
        # Get the content of the response.
        content = response['choices'][0]['message']['content']
        if verbose:
            print('CONTENT:', content)
    except:
        if verbose:
            print('OpenAI query failed.')

    # Get the structured the data.
    try:
        start_index = content.find('{')
        end_index = content.rfind('}') + 1
        extract = json.loads(content[start_index:end_index])
    except:
        extract = {}
        if verbose:
            print('JSON parsing failed.')

    # Calculate the cost of the request.
    try:
        cost = get_tokens_price(messages, model=model)
    except:
        cost = 0

    # Return the extracted data.
    return extract, cost



"""
Certificate of Analysis (CoA) Generation | Cannlytics
Copyright (c) 2021-2022 Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 2/6/2021
Updated: 1/10/2022
License: MIT LIcense <https://opensource.org/licenses/MIT>

TODO:
    - Allow users to create CoA's with a myriad of templates! (.docx, .xlsx, etc.)
    - Import any docx template styed with Jinga-style formatting and let'r'rip.
    - Use your own templates or download the started templates and customize
    them to your heart's galore.
"""
# Standard imports.
import os
from pathlib import Path
from shutil import copyfile

# External packages.
from dotenv import dotenv_values
import openpyxl
# from openpyxl.drawing.image import Image
import pandas as pd
# import qrcode

# Internal imports.
from ..firebase import (
    get_document,
    # update_document,
    download_file,
)
from ..utils.utils import snake_case


# TODO: This file needs a MAJOR refactor.


def create_coa():
    """
    Creates a certificate of analysis.
    """

    # TODO: Get sample details.

    # Get project details.

    # Get client details.

    # Get template.

    # Fill template
        # - If all of the analyses in a table have the same units,
        # then put the units in a footnote beneath. Otherwise,
        # append the units to the result.

        # Wishlist: Add initials of analyst.
        # Wishlist: Add date and time of analysis.
        # Wishlist: Add SoP, method, instrument.
        # Wishlist: Option to add dilution factor to CoA.

    # Create PDF in /tmp

    # Upload PDF to storage

    # Create short link (preferably create before upload?)

    # Create QR code

    # Update PDF, re-upload.

    return NotImplementedError


def delete_coa():
    """
    Delete a certificate of analysis from storage.
    """
    # Remove certificate from storage.

    # delete_file(bucket_name, blob_name)

    # Remove the certificate data.

    # update_document(ref, {'coa_url': '', 'coa_ref': ''})

    return NotImplementedError


def email_coas():
    """
    Email certificates of analysis to their recipients.
    """
    # Get the certificate data.

    # Email links (optional attachments) to the contacts.

    return NotImplementedError


def text_coas():
    """
    Text certificates of analysis to their recipients.
    """
    # Get the certificate data.

    # Text links to the contacts.

    return NotImplementedError


def send_coas():
    """
    Email and/or text certificates of analysis to their recipients.
    """
    # Email and / or text CoAs using email_coas and text_coas.

    return NotImplementedError


def get_coa_urls():
    """
    Get CoA URLs for given samples or projects.
    """
    # Get certificate data and return the short links.

    return NotImplementedError


def merge_coas():
    """
    Merge two CoAs together, either existing templates
    or existing PDFs.
    """
    # Get certificate data for the desired CoAs.

    # Download the existing PDFs for the certificates.

    # Merge the downloaded PDFs.

    # Upload the merged PDF to storage.

    # Create short link.

    # Create QR code.

    # Re-upload the PDF.

    return NotImplementedError



def approve_coa():
    """
    Creates a certificate of analysis.
    """
    # Verify the user's pin.

    # Get the user's signature.

    # Insert the user's signature on the certificate.

    print('Approving CoA...')
    return NotImplementedError


def review_coa(env_file='.env', signature_dest='./tmp/signature.png'):
    """
    Creates a certificate of analysis.
    """
    # TODO: Validate the user's pin to get their UID.
    uid = ''

    # Get the user's signature (if not already downloaded?).
    env_variables = dotenv_values(env_file)
    bucket_name = env_variables['FIREBASE_STORAGE_BUCKET']
    signature_data = get_document(f'users/{uid}/user_settings/signature')
    download_file(signature_data['signature_ref'], signature_dest, bucket_name)

    # Insert the signature into the CoA template.

    # Create the PDF.

    # Upload the PDF to storage.

    # Save the reviewer data.

    print('Reviewing CoA...')
    return NotImplementedError


def upload_coa():
    """
    Upload a certificate of analysis to storage.
    """
    # Read file data.

    # Upload file data to storage.

    # Update certificate data.

    return NotImplementedError


def upload_coa_template():
    """
    Upload a certificate of analysis template to storage.
    """
    # Read in the file data.

    # Upload the template to storage.

    # Create a download link for the template.

    # Save the template data.

    return NotImplementedError


def create_short_url():
    """
    Create a short URL for a CoA link.
    """
    return NotImplementedError


# def insert_qr_code(sheet, coords, url):
#     """Insert a QR code into a CoA template.
#     Args:
#         sheet (Worksheet): The worksheet to insert the QR Code
#         ref (str): The location to insert the QR code.
#         url (str): The URL that the QR code should link.
#     Returns:
#     """
#     qr = qrcode.QRCode(
#         version=1,
#         error_correction=qrcode.constants.ERROR_CORRECT_L,
#         box_size=10,
#         border=4,
#     )
#     qr.add_data(url)
#     qr.make(fit=True)
#     img = qr.make_image(fill_color='black', back_color='white')
#     img.save('qr_code.png')
#     logo = Image('qr_code.png')
#     logo.height = 150
#     logo.width = 150
#     sheet.add_image(logo, coords)
#     # workbook.save(filename="hello_world_logo.xlsx")

#-----------------------------------------------------------------------
# FIXME: Generate CoA routine
#-----------------------------------------------------------------------

def generate_coas(
        import_files,
        output_pages,
        coa_template='./coa_template.xlsm',
        # render_file='./CoAs/coa_render.xlsm',
        limits={}
):
    """Generate certificates of analysis.
    Args:
        import_files (list): A list of files to import.
        output_pages (list): A list of pages to include in the PDF.
        coa_template (str): The path of the CoA Template.
        limits (dict): A dictionary of limits and LOQ for analytes.
    """

    # Create CoA folder if one does not exist.
    Path('CoAs').mkdir(parents=True, exist_ok=True)
    dir_path = os.path.dirname(os.path.realpath(__file__))

    # Create a copy of the template.
    abs_coa_template = os.path.join(dir_path, coa_template)
    coa_template_copy = abs_coa_template.replace('.xlsm', '_copy.xlsm')
    copyfile(abs_coa_template, coa_template_copy)

    # Iterate over all import files.
    for import_file in import_files:

        # Get all sample, results, client, etc. data.
        # abs_import_file = os.path.join(dir_path, import_file)
        all_data = pd.read_excel(import_file)

        # Get all the masses.
        masses = {}
        for _, row in all_data.iterrows():
            key = snake_case(row['assay'])
            masses[key] = row['test_mass']

        # Aggregate all data for a sample
        data = all_data.groupby('sample_id', as_index=False).first()

        # Fill in sample details.
        for _, row in data.iterrows():

            # Get sample data as a dictionary.
            sample_data = {**row.to_dict(), **limits}
            sample_id = sample_data['sample_id']
            if not sample_id: # FIXME: Skip nan
                continue

            # Calculate terpene and cannabinoid results.
            sample_data = calculate_results(
                sample_data,
                analysis='terpenes',
                mass=masses['terpenes'],
                dilution_factor=40
            )
            sample_data = calculate_results(
                sample_data,
                analysis='cannabinoids',
                mass=masses['potency'],
                dilution_factor=40*50
            )

            # Iterate over worksheet pages to fill-in
            # all occurrences of each reference 1 by 1.
            template_workbook = openpyxl.load_workbook(coa_template_copy, keep_vba=True)
            fill_jinja_references(template_workbook, sample_data)

            # FIXME: Get output pages dynamically
            try:
                ws_index_list = get_worksheet_indexes(template_workbook, output_pages)
            except:
                if len(output_pages) == 3:
                    ws_index_list = [3, 4, 5]
                else:
                    ws_index_list = [3, 4, 5, 6]

            # Save the rendered template, temporarily.
            abs_render_file = os.path.join(dir_path, f'CoAs/{sample_id}.xlsm')
            template_workbook.save(abs_render_file)

            # Future: Insert signatures
            # Future: Insert QR Code
            # Future: Touch up the CoA.
                # ws.oddHeader.left.text = "Page &[Page] of &N"
                # Mark failures as red
                # a1.font = Font(color="FF0000", italic=True) # the change only affects A1

            # Create a PDF.
            output_file = f'CoAs/{sample_id}.pdf'
            # FIXME:
            excel = create_coa_pdfs(abs_render_file, ws_index_list, output_file)

            # Future: Upload the PDF.        
            # Future: Create download link and short link for the CoA.
            # Future: Upload the CoA data.
                # - See https://cloud.google.com/functions/docs/writing/http#uploading_files_via_cloud_storage

    # Remove temporary files.
    # os.remove(abs_render_file)
    os.remove(coa_template_copy)

    # Ensure Excel is visible.
    excel.ScreenUpdating = True
    excel.DisplayAlerts = True
    excel.EnableEvents = True


# FIXME: This requires that Excel is installed on Windows. Try to do this without Excel :(
def create_coa_pdfs(render_file, ws_index_list, output_file, tight=False):
    """Generate PDFs for rendred CoAs.
    Args:
        render_file (str): The path of the rendred workbook.
        ws_index_list (list): A list of the worksheet indexes to include in the PDF.
        output_file (str): The name of the output PDF (expected .pdf extension).
        tight (bool): Optional, default False, choice to scale all pages.
    Returns:
        (COM Object): A Microsoft Excel Client to alter later.
    """
    client = win32com.client.Dispatch('Excel.Application')
    client.Visible = False
    client.ScreenUpdating = False
    client.DisplayAlerts = False
    client.EnableEvents = False
    dir_path = os.path.dirname(os.path.realpath(__file__))
    input_file = os.path.join(dir_path, render_file)
    pdf_file = os.path.join(dir_path, output_file)
    workbook = client.Workbooks.Open(input_file)
    if tight:
        print_area = 'A1:G50'
        for index in ws_index_list:
            worksheet = workbook.Worksheets[index]
            worksheet.PageSetup.Zoom = False
            worksheet.PageSetup.FitToPagesTall = 1
            worksheet.PageSetup.FitToPagesWide = 1
            worksheet.PageSetup.PrintArea = print_area
    try:
        workbook.WorkSheets(ws_index_list).Select()
    except:
        workbook.Worksheets(ws_index_list).Select()
    workbook.ActiveSheet.ExportAsFixedFormat(0, pdf_file)
    workbook.Close(False)
    return client

#-----------------------------------------------------------------------
# Internal functions
#-----------------------------------------------------------------------

def fill_jinja_references(workbook, data):
    """Fill-in Jinga-style references, iterating over all
    worksheet pages to fill-in all occurrences of each reference 1 by 1.
    Args:
        workbook (Workbook): A Workbook object.
        data (dict): A dictionary of context data.
    """
    context = { snake_case(key): values for key, values in data.items() }
    sheets = workbook.worksheets
    for sheet in sheets:
        refs = get_jinja_references(sheet)
        for key, cells in refs.items():
            clean_key = snake_case(key)
            value = context.get(clean_key, 0)
            for cell in cells:
                if value:
                    sheet[cell] = str(value)
                else:
                    sheet[cell] = 'MISSING'


def get_jinja_references(sheet):
    """Get Jinja-style references in an Excel worksheet.
    Args:
        sheet (Worksheet): A worksheet to get references from.
    Returns:
        (dict): A dictionary of variables to find references.
    """
    refs = {}
    for row in sheet.iter_rows():
        for cell in row:
            value = cell.value
            try:
                if value.startswith('{{'):
                    ref = cell.column_letter + str(cell.row)
                    variable = value.replace('{{', '').replace('}}', '').strip()
                    existing_refs = refs.get(variable, [])
                    refs[variable] = [*existing_refs, *[ref]]
                elif value.startswith('{% for'): # Optional: Handle tables
                    return NotImplementedError
            except AttributeError:
                continue
    return refs


def get_worksheet_data(sheet, headers):
    """Get the data of a worksheet.
    Args:
        sheet (Worksheet): An openpyx; Excel file object.
        headres (list): A list of headers to map the values.
    Returns:
        list(dict): A list of dictionaries.
    """
    data = []
    for row in sheet.iter_rows():
        values = {}
        for key, cell in zip(headers, row):
            values[key] = cell.value
        data.append(values)
    return data


def get_worksheet_data_block(sheet, coords, expand=None):
    """Get a data block.
    Args:
        sheet (Sheet): The worksheet containing the data block.
        coords (str): The inclusive coordinates of the data block.
        expand (str): Optionally expand the range of values.
    Returns
        (dict): A dictionary of the data in the data block.
    """
    data = {}
    values = sheet.range(coords).options(expand=expand).value
    for item in values:
        key = snake_case(item[0])
        value = item[1]
        data[key] = value
    return data


def get_worksheet_headers(sheet):
    """Get the headres of a worksheet.
    Args:
        sheet (Worksheet): An openpyx; Excel file object.
    Returns:
        headers (list): A list of header strings.
    """
    headers = []
    for cell in sheet[1]:
        headers.append(cell.value)
    return headers


def get_worksheet_indexes(wb, output_pages):
    """Get the indexes for a list of sheet names in a given workbook.
    Args:
        wb (Workbook): The workbook at hand.
        pages (list): A list of pages to find indexes.
    """
    ws_index_list = []
    for page in output_pages:
        index = wb.worksheets.index(wb[page])
        ws_index_list.append(index + 1)
    return ws_index_list


if __name__ == '__main__':

    import argparse

    # Declare command line arguments.
    parser = argparse.ArgumentParser(description='CoA Generation')
    parser.add_argument('--files', action='store', dest='files', default=['export.xlsx'])
    parser.add_argument('--pages', action='store', dest='pages', default=['Page 1', 'Page 2', 'Page 3', 'Page 4'])
    parser.add_argument('--template', action='store', dest='template', default='coa_template.xlsm')
    args = parser.parse_args()

    # Format files as a list.
    files = args.files.replace('[', '').replace(']', '').split(',')
    files = [x.strip() for x in files]

    # Format pages as a list.
    pages = args.files.replace('[', '').replace(']', '').split(',')
    pages = [x.strip() for x in pages]

    # Read in analyte limits.
    limits = get_analyte_limits()

    # TEST:
    # files = './Export Files'
    # pages = ['Page 1', 'Page 2', 'Page 3', 'Page 4']

    # Optionally specify a folder instead of a list of files.
    # If its a directory, get all .xlsx files as import files.
    try:
        # dir_path = os.path.dirname(os.path.realpath(__file__))
        # full_file = os.path.join(dir_path, files)
        potential_dir = files[0]
        if os.path.exists(os.path.dirname(potential_dir)):
            import_files = []
            for file in os.listdir(potential_dir):
                if file.endswith('.xlsx'):
                    filename = os.path.join(potential_dir, file)
                    import_files.append(filename)
            files = import_files
    except TypeError:
        pass


    # Optional: Enter image files.

    # Test CoA generation.
    generate_coas(
        files,
        output_pages=pages,
        # coa_template=args.template,
        limits=limits
    )



"""
Inventory Type Model | Cannlytics
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/5/2021
Updated: 11/5/2021
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description: Inventory type data model.
"""
# Standard imports.
from dataclasses import dataclass
from datetime import datetime

# Internal imports.
from .base import Model


@dataclass
class InventoryType(Model):
    """The type for any physical inventory item."""
    _collection = 'organizations/%s/inventory_types'
    admin_method: str = ''
    brand: str = ''
    name: str = ''
    category_type: str = ''
    quantity_type: str = ''
    strain_required: bool = False
    #  "Name": "Buds",
    # "ProductCategoryType": "Buds",
    # "QuantityType": "WeightBased",
    # "RequiresStrain": true,
    # "RequiresItemBrand": false,
    # "RequiresAdministrationMethod": false,
    # "RequiresUnitCbdPercent": false,
    # "RequiresUnitCbdContent": false,
    # "RequiresUnitCbdContentDose": false,
    # "RequiresUnitThcPercent": false,
    # "RequiresUnitThcContent": false,
    # "RequiresUnitThcContentDose": false,
    # "RequiresUnitVolume": false,
    # "RequiresUnitWeight": false,
    # "RequiresServingSize": false,
    # "RequiresSupplyDurationDays": false,
    # "RequiresNumberOfDoses": false,
    # "RequiresPublicIngredients": false,
    # "RequiresDescription": false,
    # "RequiresProductPhotos": 0,
    # "RequiresLabelPhotos": 0,
    # "RequiresPackagingPhotos": 0,
    # "CanContainSeeds": true,
    # "CanBeRemediated": true



"""
Cannlytics Data Initialization | Cannlytics
Copyright (c) 2023 Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 6/15/2023
Updated: 7/4/2023
"""
from .ai import (
    initialize_openai,
    estimate_tokens_of_messages,
    estimate_tokens_of_string,
    get_prompt_price,
    get_messages_price,
    get_tokens_price,
    split_string,
    split_into_token_chunks,
    gpt_to_json,
    AI_WARNING,
    INSTRUCTIONAL_PROMPT,
    PRICE_PER_1000_TOKENS,
    MAX_PROMPT_LENGTH,
)

__all__ = [
    initialize_openai,
    estimate_tokens_of_messages,
    estimate_tokens_of_string,
    get_prompt_price,
    get_messages_price,
    get_tokens_price,
    split_string,
    split_into_token_chunks,
    gpt_to_json,
    AI_WARNING,
    INSTRUCTIONAL_PROMPT,
    PRICE_PER_1000_TOKENS,
    MAX_PROMPT_LENGTH,
]



"""
BudSpender | Cannabis Receipt Parser
Copyright (c) 2023 Cannlytics

Authors:
    Keegan Skeate <https://github.com/keeganskeate>
    Candace O'Sullivan-Sutherland <https://github.com/candy-o>
Created: 4/24/2023
Updated: 7/2/2023
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>
"""
# Standard imports:
from datetime import datetime
import json
import gc
import re
from typing import List, Optional

# External imports:
from cannlytics import __version__
from cannlytics.ai import (
    AI_WARNING,
    INSTRUCTIONAL_PROMPT,
    MAX_PROMPT_LENGTH,
    gpt_to_json,
)
from cannlytics.data import create_hash
import cv2
import pandas as pd
import pdfplumber
import pytesseract


# Prompt to parse data from a cannabis receipt.
RECEIPTS_PROMPT = """Given text, extract JSON, where:

| Field | Example | Description |
|-------|---------|-------------|
| `date_sold` | "2020-04-20" | The date the receipt was sold. |
| `invoice_number` | "123456789" | The receipt number. |
| `product_names` | ["Blue Rhino Pre-Roll"] | The names of the product purchased. |
| `product_types` | ["flower"] | The types of the products purchased. |
| `product_quantities` | [1] | The quantities of the products purchased. |
| `product_prices` | [5.0] | The prices of the products purchased. |
| `product_ids` | ["5f8b9c4b0f5c4b0008d1b2b0"] | The IDs of the products purchased. |
| `total_amount` | 5.0 | The total amount of all product prices. |
| `subtotal` | 5.0 | The subtotal of the receipt. |
| `total_discount` | 0.0 | The amount of discount applied to the transaction, if applicable. |
| `total_paid` | 5.0 | The total amount paid. |
| `change_due` | 0.0 | The amount of change due. |
| `rewards_earned` | 0.0 | The amount of rewards earned. |
| `rewards_spent` | 0.0 | The amount of rewards spent. |
| `total_rewards` | 0.0 | The total amount of rewards. |
| `city_tax` | 0.0 | The amount of city tax applied to the transaction, if applicable. |
| `county_tax` | 0.0 | The amount of county tax applied to the transaction, if applicable. |
| `state_tax` | 0.0 | The amount of state tax applied to the transaction, if applicable. |
| `excise_tax` | 0.0 | The amount of excise tax applied to the transaction, if applicable. |
| `retailer` | "BudHouse" | The name of the retailer. |
| `retailer_license_number` | "C11-0000001-LIC" | The license number of the retailer. |
| `retailer_address` | "1234 Main St, San Diego, CA 92101" | The address of the retailer. |
| `retailer_street` | "420 State Ave" | The retailer street, if applicable. |
| `retailer_city` | "Olympia" | The retailer city, if applicable. |
| `retailer_state` | "CA" | The state of the retailer, if applicable. |
| `retailer_zipcode` | "98506" | The zip code of the retailer, if applicable. |
| `budtender` | "John Doe" | The name of the budtender. |
"""


class ReceiptsParser(object):
    """A cannabis receipt parser, powered by OpenAI."""

    def __init__(
            self,
            default_config: Optional[str] = '--oem 3 --psm 6',
            openai_api_key: Optional[str] = None,
            model: Optional[str] = 'gpt-4',
            base_prompt: Optional[str] = RECEIPTS_PROMPT,
            instructions: Optional[str] = INSTRUCTIONAL_PROMPT,
            max_tokens: Optional[int] = MAX_PROMPT_LENGTH,
        ) -> None:
        """Initialize an Open Data API client.
        Args:
            default_config: The default configuration for OpenCV.
            openai_api_key: The OpenAI API key.
            model: The OpenAI model to use, 'gpt-4' by default.
            base_prompt: The base prompt to use for the OpenAI model.
        """
        # Parameters.
        self.default_config = default_config
        self.openai_api_key = openai_api_key
        self.model = model
        self.base_prompt = base_prompt
        self.instructions = instructions
        self.max_tokens = max_tokens

        # State.
        self.total_cost = 0
        self.img = None
        self.image_text = ''
        self.extracted_data = None
    
    def image_to_pdf_to_text(self, image_file: str) -> str:
        """Extract the text from an image by converting it to a PDF.
        Args:
            image_file: The path to the image file.
        """
        img = self.rotate(cv2.imread(str(image_file), cv2.IMREAD_COLOR))
        pdf = pytesseract.image_to_pdf_or_hocr(img, extension='pdf')
        pdf_file = image_file.replace('.jpg', '.pdf')
        with open(pdf_file, 'w+b') as f:
            f.write(pdf)
            f.close()
        report = pdfplumber.open(pdf_file)
        text = ''
        for page in report.pages:
            text += page.extract_text()
        return text
    
    def image_to_text(
            self,
            image_file: str,
            median_blur: Optional[int] = 0
        ) -> str:
        """Extract the text from an image.
        Args:
            image_file: The path to the image file.
            median_blur: Removes noise. Must be a positive odd integer.
        """
        img = self.rotate(cv2.imread(str(image_file), cv2.IMREAD_COLOR))
        img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
        if median_blur: img = cv2.medianBlur(img, median_blur)
        text = pytesseract.image_to_string(img, config=self.default_config)
        return text

    def rotate(self, image, center=None, scale=1.0):
        """Rotate an image with text into a readable position, as well as possible.
        Credit: Mousam Singh <https://stackoverflow.com/a/55122911/5021266>
        License: CC BY-SA 4.0 <https://creativecommons.org/licenses/by-sa/4.0/>
        """
        (h, w) = image.shape[:2]
        if center is None:
            center = (w / 2, h / 2)
        degrees = int(re.search(
            '(?<=Rotate: )\d+',
            pytesseract.image_to_osd(image),
        ).group(0))
        angle = 360 - degrees
        M = cv2.getRotationMatrix2D(center, angle, scale)
        rotated = cv2.warpAffine(image, M, (w, h))
        return rotated
    
    def parse(
            self,
            doc: str,
            openai_api_key: Optional[str] = None,
            model: Optional[str] = 'gpt-4',
            max_tokens: Optional[int] = None,
            temperature: Optional[float] = 0.0,
            system_prompts: Optional[List[str]] = None,
            verbose: Optional[bool] = False,
            user: Optional[str] = 'cannlytics',
            retry_pause: Optional[float] = 3.33,
        ) -> dict:
        """Parse a receipt with OpenAI's GPT model and return the data as JSON."""
        # Initialize parameters.
        if max_tokens is None:
            max_tokens = self.max_tokens

        # Get the text of a receipt.
        text = self.image_to_text(doc)

        # Begin the observation with the hash of the text.
        obs = {}
        obs['hash'] = create_hash(text, private_key='')

        # Format the system prompts.
        if not system_prompts:
            system_prompts = [self.instructions, self.base_prompt]

        # Initialize OpenAI.
        if openai_api_key is None:
            openai_api_key = self.openai_api_key
        if model is None:
            model = self.model

        # Parse the receipt.
        extracted_data, cost = gpt_to_json(
            text,
            system_prompts=system_prompts,
            model=model,
            max_tokens=max_tokens,
            temperature=temperature,
            user=user,
            openai_api_key=openai_api_key,
            verbose=verbose,
            retry_pause=retry_pause,
        )
        obs = {**obs, **extracted_data}

        # Calculate total price.
        prices = obs.get('product_prices', [])
        obs['total_price'] = 0
        for price in prices:
            try:
                obs['total_price'] += float(price)
            except:
                pass

        # Calculate total tax.
        obs['total_tax'] = 0
        tax_keys = [key for key in obs.keys() if '_tax' in key]
        for key in tax_keys:
            try:
                obs['total_tax'] += float(obs[key])
            except:
                pass

        # Calculate total transactions.
        obs['total_transactions'] = len(prices)

        # Future work: Augment retailer data.

        # Future work: Augment product data with product ID.

        # Future work: Save a copy of the image to Firebase Storage.

        # Mint the observation with unique IDs.
        obs['algorithm'] = 'receipts_ai.py'
        obs['algorithm_entry_point'] = 'parse_receipt_with_ai'
        obs['algorithm_version'] = __version__
        obs['parsed_at'] = datetime.now().isoformat()
        obs['warning'] = AI_WARNING
        self.total_cost += cost
        return obs
    
    def save(self, obs, filename):
        """Save a receipt to a file."""
        if filename.endswith('json'):
            with open(filename, 'w') as f:
                json.dump(obs, f, indent=4)
                f.close()
        elif filename.endswith('csv'):
            pd.DataFrame([obs]).to_csv(filename)
        elif filename.endswith('xlsx'):
            pd.DataFrame([obs]).to_excel(filename)
        else:
            raise ValueError(f'Unknown file type: {filename}')
        
    def quit(self):
        """Reset the parser and perform garbage cleaning."""
        self.model = 'gpt-4'
        self.openai_api_key = None
        self.total_cost = 0
        self.img = None
        self.image_text = ''
        self.extracted_data = None
        gc.collect()


# === Tests ===
if __name__ == '__main__':

    # Initialize a receipt parser.

    # Initialize OpenAI.
    from dotenv import dotenv_values
    config = dotenv_values('../../../.env')
    openai_api_key = config['OPENAI_API_KEY']

    # Initialize a receipt parser.
    parser = ReceiptsParser(
        model='gpt-4',
        openai_api_key=openai_api_key,
    )

    # [✓] TEST: Parse a receipt.
    image_file = '../../../.datasets/receipts/tests/receipt-1.jpg'
    outfile = '../../../.datasets/receipts/tests/receipt-1.xlsx'
    receipt_data = parser.parse(image_file)
    print('Cost:', parser.total_cost)
    parser.save(receipt_data, outfile)

    # # [ ] TEST: Parse a folder of receipts.
    # RECEIPT_DIR = '../../../.datasets/receipts/tests/'
    # all_receipt_data = []
    # file_types = ['.jpg', '.png', '.jpeg']
    # filenames = os.listdir(RECEIPT_DIR)
    # for filename in filenames:
    #     if any(filename.endswith(file_type) for file_type in file_types):
    #         image_file = os.path.join(RECEIPT_DIR, filename)
    #         receipt_data = parser.parse(image_file)
    #         all_receipt_data.append(receipt_data)



"""
Transfer Management | Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>  
Created: 6/23/2021  
Updated: 6/23/2021  
License: MIT License <https://opensource.org/licenses/MIT>  

Tools to help manage transfers of laboratory samples.
"""
# TODO: Does this belong in crm?

def create_transfer():
    """Create a transfer of samples to a lab for analysis. The sending
    organization specifies the sample details and the analyses requested.
    The receiving organization, the lab, can create projects from the
    sample details and analyses requested. The lab can edit sample details
    stored in a project, but only the sender can edit sample details in
    the transfer. By default, the sending organization will receive
    results when they are released."""

    return NotImplementedError


def receive_transfer(transfer_id: str):
    """
    Receive a transfer of laboratory samples.
    """
    packages = track.get_packages(license_number=cultivator.license_number)
    return NotImplementedError


def reject_transfer():
    """
    Reject a transfer of laboratory samples.
    """
    return NotImplementedError




"""
User Model | Cannlytics
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/5/2021
Updated: 11/5/2021
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description: User data model.
"""
# Standard imports.
from dataclasses import dataclass
from datetime import datetime

# Internal imports.
from .base import Model


@dataclass
class User(Model):
    """The person performing the work."""
    _collection = 'users/%s'
    email: str = ''
    license_number: str = ''
    linked_in_url: str = ''
    name: str = ''
    phone_number: str = ''
    phone_number_formatted: str = ''
    signature_storage_ref: str = ''
    signed_in: bool = False
    signed_in_at: datetime = None



"""
Result Calculation | Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>  
Created: 6/23/2021  
Updated: 8/14/2021  
License: MIT License <https://opensource.org/licenses/MIT>  

Use analyte limits and formulas and instrument measurements to calculate
final results for analyses.
"""
# Standard imports.
from datetime import datetime
from email.mime.multipart import MIMEMultipart
from smtplib import SMTP

# Internal imports.
from cannlytics.firebase import get_collection, update_documents
from cannlytics.utils import encode_pdf


def calculate_results(sample_data, analysis, mass, dilution_factor=10, correction_factor=10000):
    """Calculate percentage results given raw results,
    dilution factor, and analysis type.
    Args:
        sample_data (dict): A dictionary of sample data.
        analysis (str): An analysis to calculate results for the analysis's analytes.
        mass (float): The recorded sample mass.
        dilution_factor (float): The dilution factor for the sample.
        correction_factor (float): A factor used to appropriately scale values to percentage.
    Returns:
        (dict): An updated dictionary of sample data.
    """
    # FIXME:
    # analytes = get_analytes(analysis)
    analytes = []
    for analyte in analytes:
        try:
            raw_value = float(sample_data[analyte])
            sample_data[analyte] = ((raw_value * dilution_factor) / mass) / correction_factor
        except ValueError:
            continue
    return sample_data


# TODO: Implement.
def convert_results_units():
    """Convert results between units."""
    raise NotImplementedError


def post_results():
    """
    Post results to your state traceability system.
    """
    # # Initialize Firebase.
    # config = dotenv_values('../../../.env')
    # os.environ['GOOGLE_APPLICATION_CREDENTIALS'] = config['GOOGLE_APPLICATION_CREDENTIALS']
    # db = fb.initialize_firebase()

    # # Initialize a Metrc client.
    # vendor_api_key = config['METRC_TEST_VENDOR_API_KEY']
    # user_api_key = config['METRC_TEST_USER_API_KEY']
    # track = metrc.authorize(vendor_api_key, user_api_key)

    # # Upload PDF.
    # encoded_pdf = encode_pdf('../assets/pdfs/example_coa.pdf')
    
    # # Upload lab results
    # lab_result_data = {
    #     'Label': test_package_label,
    #     'ResultDate': get_timestamp(),
    #     # 'LabTestDocument': {
    #         # 'DocumentFileName': 'new-old-time-moonshine.pdf',
    #         # 'DocumentFileBase64': 'encoded_pdf',
    #     # },
    #     'Results': [
    #         {
    #             'LabTestTypeName': 'THC',
    #             'Quantity': 0.07,
    #             'Passed': True,
    #             'Notes': ''
    #         },
    #         {
    #             'LabTestTypeName': 'CBD',
    #             'Quantity': 23.33,
    #             'Passed': True,
    #             'Notes': ''
    #         },
    #         # {
    #         #     'LabTestTypeName': 'Microbiologicals',
    #         #     'Quantity': 0,
    #         #     'Passed': True,
    #         #     'Notes': ''
    #         # },
    #         # {
    #         #     'LabTestTypeName': 'Pesticides',
    #         #     'Quantity': 0,
    #         #     'Passed': True,
    #         #     'Notes': ''
    #         # },
    #         # {
    #         #     'LabTestTypeName': 'Heavy Metals',
    #         #     'Quantity': 0,
    #         #     'Passed': True,
    #         #     'Notes': ''
    #         # },
    #     ]
    # }
    # track.post_lab_results([lab_result_data], license_number=lab.license_number)
    
    # Get tested package. (background)
    # test_package = track.get_packages(label=test_package_label, license_number=lab.license_number)

    # Get the tested package's lab result. (background)
    # lab_results = track.get_lab_results(uid=test_package.id, license_number=lab.license_number)
    
    return NotImplementedError


def release_results(org_id, sample_ids, released_by=None):
    """
    Release completed results to their recipients,
    making them available through the client portal
    and the state traceability system.
    Args:
        org_id (str): The ID of the organization releasing results.
        sample_ids (list): A list of sample IDs to release.
        released_by (str): Optional ID of the user who released the sample results.
    """
    refs = []
    data = []
    for sample_id in sample_ids:
        refs.append(f'organizations/{org_id}/samples/{sample_id}')
        data.append({
            'released': True,
            'released_at': datetime.now().isoformat(),
            'released_by': released_by,
        })
    update_documents(refs, data)


def email_results(samples, recipients):
    """Email results to their recipients.
    Args:
        samples (list): A list of sample data (dict).
        recipients (list): A list of recipient emails.
    Returns:
        (list): A list of success or fail indicators (bool).
    """

    # TODO: Format HTML message.

    # TODO: Write email logic.
    # rcpt = cc.split(",") + bcc.split(",") + [to]
    # msg = MIMEMultipart('alternative')
    # msg['Subject'] = "my subject"
    # msg['To'] = to
    # msg['Cc'] = cc
    # msg.attach(my_msg_body)
    # server = SMTP("localhost") # or your smtp server
    # server.sendmail(me, rcpt, msg.as_string())
    # server.quit()
    return NotImplementedError


def text_results(samples, recipients):
    """
    Text results to their recipients.
    Args:
        samples (list): A list of sample data (dict).
        recipients (list): A list of recipient emails.
    Returns:
        (list): A list of success or fail indicators (bool).
    """
    # TODO: Get sending email, password, sending address.
    server = SMTP('smtp.gmail.com', 587 )
    server.starttls()
    server.login('xxxxx@gmail.com', 'xxxxxxxxxx')
    from_mail = 'xxxxxxxxx@gmail.com'

    # TODO: Format message
    body = ''
    for sample in samples:
        body += '%s: %s\n' % (sample['sample_id'], sample['coa_short_url'])

    # TODO: Send text message to each recipient.
    successes = []
    for recipient in recipients:
        try:
            to = '9xxxxxxx@tmomail.net'
            message = (
                'From: %s\r\n' % from_mail
                + 'To: %s\r\n' % to
                # + "CC: %s\r\n" % ",".join(cc)
                + 'Subject: %s\r\n' % ''
                + '\r\n' + body)
            server.sendmail(from_mail, to, message)
            successes.append(True)
        except:
            successes.append(False)
    return successes


def send_results(org_id, sample_ids, recipients, method='email'):
    """Send results to their recipients with email or text message.
    Sample data is retrieved in batches of 10 to use for text or email.
    Args:
        org_id (str): The organization sending the results.
        sample_ids (list): A list of sample IDs for which to send results.
        recipients (list): A list of recipients, either a list of emails or
            phone numbers
        method (str): The method of result delivery, `email` or `text`,
            with `email` by default.
    Returns:
        (list): A list of all success indicators (bool).
    """
    samples = []
    batches = [sample_ids[i:i+10] for i in range(0, len(sample_ids), 10)]
    for batch in batches:
        filters = [{'key': 'sample_id', 'operation': 'in', 'value': batch}]
        docs = get_collection(f'organizations/{org_id}/samples', filters=filters)
        samples = [*samples, *docs]
    if method == 'text':
        successes = text_results(samples, recipients)
    else:
        successes = email_results(samples, recipients)
    return successes
